#!/usr/bin/env python
# -*- coding: utf-8 -*-

import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import axes3d
from scipy import signal

import spectral_utils
import separation_base
import audio_signal
import constants


class Duet(separation_base.SeparationBase):
    """Implements the Degenerate Unmixing Estimation Technique (DUET) algorithm.

    The DUET algorithm was originally proposed by S.Rickard and F.Dietrich for DOA estimation
    and further developed for BSS and demixing by A.Jourjine, S.Rickard, and O. Yilmaz.


    References:

        * Rickard, Scott. "The DUET blind source separation algorithm." Blind Speech Separation. Springer Netherlands,
          2007. 217-241.
        * Yilmaz, Ozgur, and Scott Rickard. "Blind separation of speech mixtures via time-frequency masking." Signal
          Processing, IEEE transactions on 52.7 (2004): 1830-1847.

    Parameters:
        audio_signal (np.array): a 2-row Numpy matrix containing samples of the two-channel mixture
        num_sources (int): number of sources to find
        a_min (Optional[int]): Minimum attenuation. Defaults to -3
        a_max(Optional[int]): Maximum attenuation. Defaults to 3
        a_num(Optional[int]): Number of bins for attenuation. Defaults to 50
        d_min (Optional[int]): Minimum delay. Defaults to -3
        d_max (Optional[int]): Maximum delay. Defaults to 3
        d_num (Optional[int]): Number of bins for delay. Defaults to 50
        threshold (Optional[float]): Value in [0, 1] for peak picking. Defaults to 0.2
        a_min_distance (Optional[int]): Minimum distance between peaks wrt attenuation. Defaults to 5
        d_min_distance (Optional[int]): Minimum distance between peaks wrt delay. Defaults to 5
        stft_params (Optional[WindowAttributes]): Window attributes for stft. Defaults to
         WindowAttributes.WindowAttributes(self.sample_rate)
        sample_rate (Optional[int]): Sample rate for the audio. Defaults to Constants.DEFAULT_SAMPLE_RATE

    Examples:
        :ref:`The DUET Demo Example <duet_demo>`

    """

    def __init__(self, input_audio_signal, num_sources, sample_rate=None, stft_params=None,
                 a_min=-3, a_max=3, a_num=50, d_min=-3, d_max=3, d_num=50,
                 threshold=0.2, a_min_distance=5, d_min_distance=5):
        # TODO: Is there a better way to do this?
        self.__dict__.update(locals())
        super(Duet, self).__init__(input_audio_signal, sample_rate, stft_params)
        self.separated_sources = None
        self.a_grid = None
        self.d_grid = None
        self.hist = None

    def __str__(self):
        return 'Duet'

    def run(self):
        """Extracts N sources from a given stereo audio mixture (N sources captured via 2 sensors)

        Returns:
            * **xhat** (*np.array*) - an N-row matrix containing time-domain estimates of sources
            * **ad_est** (*np.array*) - N by 2 Numpy matrix containing estimated attenuation and delay values
              corresponding to N sources
        Example:
             ::
            input_file_name = '../Input/dev1_female3_inst_mix.wav'
            signal = AudioSignal(path_to_input_file=input_file_name)

            duet = Duet(signal, a_min=-3, a_max=3, a_num=50, d_min=-3, d_max=3, d_num=50, threshold=0.2,
            a_min_distance=5, d_min_distance=5, num_sources=3)

            duet.run()

        """

        if self.audio_signal.num_channels != 2:
            raise Exception('Cannot run DUET on audio signal without exactly 2 channels!')

        # Give them shorter names
        L = self.stft_params.window_length
        winType = self.stft_params.window_type
        hop = self.stft_params.hop_length
        fs = self.sample_rate

        # Compute the stft of the two channel mixtures
        X1, P1, F, T = spectral_utils.e_stft_plus(self.audio_signal.get_channel(1), L, hop, winType, fs)
        X2, P2, F, T = spectral_utils.e_stft_plus(self.audio_signal.get_channel(2), L, hop, winType, fs)

        # remove dc component to avoid dividing by zero freq. in the delay estimation
        X1 = X1[1::, :]
        X2 = X2[1::, :]
        Lf = len(F)
        Lt = len(T)

        # Compute the freq. matrix for later use in phase calculations
        wmat = np.array(np.tile(np.mat(F[1::]).T, (1, Lt))) * (2 * np.pi / fs)  # WTF?

        # Calculate the symmetric attenuation (alpha) and delay (delta) for each
        # time-freq. point
        R21 = (X2 + constants.EPSILON) / (X1 + constants.EPSILON)
        atn = np.abs(R21)  # relative attenuation between the two channels
        alpha = atn - 1 / atn  # symmetric attenuation
        delta = -np.imag(np.log(R21)) / (2 * np.pi * wmat)  # relative delay

        # What is going on here???
        # ------------------------------------
        # calculate the weighted histogram
        p = 1
        q = 0
        tfw = (np.abs(X1) * np.abs(X2)) ** p * (np.abs(wmat)) ** q  # time-freq weights

        # only consider time-freq. points yielding estimates in bounds
        a_premask = np.logical_and(self.a_min < alpha, alpha < self.a_max)
        d_premask = np.logical_and(self.d_min < delta, delta < self.d_max)
        ad_premask = np.logical_and(a_premask, d_premask)

        ad_nzind = np.nonzero(ad_premask)
        alpha_vec = alpha[ad_nzind]
        delta_vec = delta[ad_nzind]
        tfw_vec = tfw[ad_nzind]

        # compute the histogram
        H = np.histogram2d(alpha_vec, delta_vec, bins=np.array([self.a_num, self.d_num]),
                           range=np.array([[self.a_min, self.a_max], [self.d_min, self.d_max]]), normed=False,
                           weights=tfw_vec)

        hist = H[0] / H[0].max()
        agrid = H[1]
        dgrid = H[2]

        # Save these for later
        self.a_grid = agrid
        self.d_grid = dgrid
        self.hist = hist
        self.non_normalized_hist = H[0]

        # smooth the histogram - local average 3-by-3 neighboring bins
        hist = self.twoDsmooth(hist, np.array([3]))

        # normalize and plot the histogram
        hist /= hist.max()

        # find the location of peaks in the alpha-delta plane
        pindex = self.find_peaks2(hist, self.threshold,
                                  np.array([self.a_min_distance, self.d_min_distance]), self.num_sources)

        alphapeak = agrid[pindex[0, :]]
        deltapeak = dgrid[pindex[1, :]]

        ad_est = np.vstack([alphapeak, deltapeak]).T

        # convert alpha to a
        atnpeak = (alphapeak + np.sqrt(alphapeak ** 2 + 4)) / 2

        # compute masks for separation
        bestsofar = np.inf * np.ones((Lf - 1, Lt))
        bestind = np.zeros((Lf - 1, Lt), int)
        for i in range(0, self.num_sources):
            score = np.abs(atnpeak[i] * np.exp(-1j * wmat * deltapeak[i]) * X1 - X2) ** 2 / (1 + atnpeak[i] ** 2)
            mask = (score < bestsofar)
            bestind[mask] = i
            bestsofar[mask] = score[mask]

        # demix with ML alignment and convert to time domain
        Lx = self.audio_signal.signal_length
        xhat = np.zeros((self.num_sources, Lx))
        for i in range(0, self.num_sources):
            mask = (bestind == i)
            Xm = np.vstack([np.zeros((1, Lt)),
                            (X1 + atnpeak[i] * np.exp(1j * wmat * deltapeak[i]) * X2) / (1 + atnpeak[i] ** 2) * mask])
            # xi = spectral_utils.f_istft(Xm, L, winType, hop, fs)
            xi = spectral_utils.e_istft(Xm, L, hop, winType)

            xhat[i, :] = xi[0:Lx]
            # add back to the separated signal a portion of the mixture to eliminate
            # most of the masking artifacts
            # xhat=xhat+0.05*x[0,:]

        self.separated_sources = xhat

        return xhat, ad_est

    @staticmethod
    def find_peaks2(data, min_thr=0.5, min_dist=None, max_peaks=1):
        """Receives a matrix of positive numerical values (in [0,1]) and finds the peak values and corresponding indices.

        Parameters:
            data (np.array): a 2D Numpy matrix containing real values (in [0,1])
            min_thr (Optional[float]): minimum threshold (in [0,1]) on data values. Defaults to 0.5
            min_dist (Optional[np.array]): 1 by 2 matrix containing minimum distances (in # of time elements) between
             peaks row-wise and column-wise. Defaults to .25* matrix dimensions
            max_peaks (Optional[int]): maximum number of peaks in the whole matrix. Defaults to 1

        Returns:
            peakIndex (np.array): a two-row Numpy matrix containing peaks and their indices
        """
        assert (type(data) == np.ndarray)

        Rdata, Cdata = data.shape
        if min_dist is None:
            min_dist = np.array([np.floor(Rdata / 4), np.floor(Cdata / 4)])

        peakIndex = np.zeros((2, max_peaks), int)
        Rmd, Cmd = min_dist.astype(int)

        # keep only the values that pass the threshold
        data = np.multiply(data, (data >= min_thr))

        if np.size(np.nonzero(data)) < max_peaks:
            raise ValueError('not enough number of peaks! change parameters.')
        else:
            i = 0
            while i < max_peaks:
                peakIndex[:, i] = np.unravel_index(data.argmax(), data.shape)
                n1 = peakIndex[0, i] - Rmd - 1
                m1 = peakIndex[0, i] + Rmd + 1
                n2 = peakIndex[1, i] - Cmd - 1
                m2 = peakIndex[1, i] + Cmd + 1
                data[n1:m1, n2:m2] = 0
                i += 1
                if np.sum(data) == 0:
                    break

        return peakIndex

    @staticmethod
    def find_peaks(data, min_thr=0.5, min_dist=None, max_num=1):
        # TODO: consolidate both find_peaks functions
        # TODO: when is this used?
        """Receives a row vector of positive values (in [0,1]) and finds the peak values and corresponding indices.

       Parameters:
            data (np.array): a 2D Numpy matrix containing real values (in [0,1])
            min_thr (Optional[float]): minimum threshold (in [0,1]) on data values. Defaults to 0.5
            min_dist (Optional[np.array]): 1 by 2 matrix containing minimum distances (in # of time elements) between
             peaks row-wise and column-wise. Defaults to .25* matrix dimensions
            max_peaks (Optional[int]): maximum number of peaks in the whole matrix. Defaults to 1

        Returns:
            peakIndex (np.array): a two-row Numpy matrix containing peaks and their indices
        """
        assert (type(data) == np.ndarray)

        if min_dist is None:
            min_dist = np.floor(data.shape[1] / 4)

        peak_indices = np.zeros((1, max_num), int)

        # keep only values above the threshold
        data = np.multiply(data, (data >= min_thr))

        if np.size(np.nonzero(data)) < max_num:
            raise ValueError('not enough number of peaks! change parameters.')
        else:
            i = 0
            while i < max_num:
                peak_indices[0, i] = np.argmax(data)
                n = peak_indices[0, i] - min_dist - 1
                m = peak_indices[0, i] + min_dist + 1
                data[0, n: m] = 0
                i += 1
                if np.sum(data) == 0:
                    break

        peak_indices = np.sort(peak_indices)

        return peak_indices

    def twoDsmooth(self, Mat, Kernel):
        """Performs two-dimensional convolution in order to smooth the values of matrix elements.

        (similar to low-pass filtering)

        Parameters:
            Mat (np.array): a 2D Numpy matrix to be smoothed
            Kernel (np.array): a 2D Numpy matrix containing kernel values
        Note:
            if Kernel is of size 1 by 1 (scalar), a Kernel by Kernel matrix of 1/Kernel**2 will be used as the matrix
            averaging kernel
        Output:
            SMat (np.array): a 2D Numpy matrix containing a smoothed version of Mat (same size as Mat)
        """

        # check the dimensions of the Kernel matrix and set the values of the averaging
        # matrix, Kmat
        if np.prod(Kernel.shape) == 1:
            Kmat = np.ones((Kernel, Kernel)) / Kernel ** 2
        else:
            Kmat = Kernel

        # make Kmat have odd dimensions
        krow, kcol = np.shape(Kmat)
        if np.mod(krow, 2) == 0:
            Kmat = signal.convolve2d(Kmat, np.ones((2, 1))) / 2
            krow += 1

        if np.mod(kcol, 2) == 0:
            Kmat = signal.convolve2d(Kmat, np.ones((1, 2))) / 2
            kcol += 1

        # adjust the matrix dimension for convolution
        copyrow = int(np.floor(krow / 2))  # number of rows to copy on top and bottom
        copycol = int(np.floor(kcol / 2))  # number of columns to copy on either side

        # TODO: This is very ugly. Make this readable
        # form the augmented matrix (rows and columns added to top, bottom, and sides)
        Mat = np.mat(Mat)  # make sure Mat is a Numpy matrix
        augMat = np.vstack(
            [
                np.hstack(
                    [Mat[0, 0] * np.ones((copyrow, copycol)),
                     np.ones((copyrow, 1)) * Mat[0, :],
                     Mat[0, -1] * np.ones((copyrow, copycol))
                     ]),
                np.hstack(
                    [Mat[:, 0] * np.ones((1, copycol)),
                     Mat,
                     Mat[:, -1] * np.ones((1, copycol))]),
                np.hstack(
                    [Mat[-1, 1] * np.ones((copyrow, copycol)),
                     np.ones((copyrow, 1)) * Mat[-1, :],
                     Mat[-1, -1] * np.ones((copyrow, copycol))
                     ]
                )
            ]
        )

        # perform two-dimensional convolution between the input matrix and the kernel
        SMAT = signal.convolve2d(augMat, Kmat[::-1, ::-1], mode='valid')

        return SMAT

    def make_audio_signals(self):
        """Returns the extracted signals

        Returns:
            signals (List[AudioSignal]): List of AudioSignals extracted using DUET.

        """
        signals = []
        for i in range(self.num_sources):
            cur_signal = audio_signal.AudioSignal(audio_data_array=self.separated_sources[i],
                                                  sample_rate=self.sample_rate)
            signals.append(cur_signal)
        return signals

    def plot(self, output_name, three_d_plot=False, normalize=True):
        """Plots histograms with the results of the DUET algorithm

        Parameters:
            output_name (str): path to save plot as
            three_d_plot (Optional[bool]): Flags whether or not to plot in 3d. Defaults to False
        """
        plt.close('all')

        AA = np.tile(self.a_grid[1::], (self.d_num, 1)).T
        DD = np.tile(self.d_grid[1::].T, (self.a_num, 1))

        histogram_data = self.hist if normalize else self.non_normalized_hist

        # plot the histogram in 2D
        if not three_d_plot:
            plt.figure()
            plt.pcolormesh(AA, DD, histogram_data)
            plt.xlabel(r'$\alpha$', fontsize=16)
            plt.ylabel(r'$\delta$', fontsize=16)
            plt.title(r'$\alpha-\delta$ Histogram')
            plt.axis('tight')
            plt.savefig(output_name)
            plt.close()

        else:
            # plot the histogram in 3D
            fig = plt.figure()
            ax = fig.add_subplot(111, projection='3d')
            ax.plot_wireframe(AA, DD, histogram_data, rstride=2, cstride=2)
            plt.xlabel(r'$\alpha$', fontsize=16)
            plt.ylabel(r'$\delta$', fontsize=16)
            plt.title(r'$\alpha-\delta$ Histogram')
            plt.axis('tight')
            ax.view_init(30, 30)
            plt.savefig(output_name)
            plt.close()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import numpy as np
import scipy.fftpack as scifft
import scipy.spatial.distance

import spectral_utils
import separation_base
import constants
from audio_signal import AudioSignal


class Repet(separation_base.SeparationBase):
    """Implements the REpeating Pattern Extraction Technique algorithm using the Similarity Matrix (REPET-SIM).

    REPET is a simple method for separating the repeating background from the non-repeating foreground in a piece of
    audio mixture. REPET-SIM is a generalization of REPET, which looks for similarities instead of periodicities.

    References:

        * Zafar Rafii and Bryan Pardo. "Audio Separation System and Method," US20130064379 A1, US 13/612,413, March 14,
          2013
        * Zafar Rafii and Bryan Pardo. "Music/Voice Separation using the Similarity Matrix," 13th International Society
          on Music Information Retrieval, Porto, Portugal, October 8-12, 2012.

    See Also:
        http://music.eecs.northwestern.edu/research.php?project=repet

    Parameters:
            audio_signal (AudioSignal): audio mixture (M by N) containing M channels and N time samples
            repet_type (RepetType): Variant of Repet algorithm to perform.
            stft_params (WindowAttributes): WindowAttributes object describing the window used in the repet
             algorithm
            sample_rate (int): the sample rate of the audio signal
            similarity_threshold (Optional[int]): Used for RepetType.SIM. Defaults to 0
            min_distance_between_frames (Optional[int]): Used for RepetType.SIM. Defaults to 1
            max_repeating_frames (Optional[int]): Used for RepetType.SIM. Defaults to 10
            period (Optional[float]): Used for RepetType.ORIGINAL. The Period of the repeating part of the signal.
            min_period (Optional[float]): Used for RepetType.ORIGINAL. Only used if Period is not provided. Defaults to
             0.8
            max_period (Optional[float]): Used for RepetType.ORIGINAL. Only used if Period is not provided. Defaults to
             min(8, self.Mixture.SignalLength/3)
            high_pass_cutoff (Optional[int]): Defaults to 100

    Examples:
        :ref:`The REPET Demo Example <repet_demo>`
    """
    def __init__(self, input_audio_signal=None, sample_rate=None, stft_params=None, repet_type=None,
                 similarity_threshold=None, min_distance_between_frames=None, max_repeating_frames=None,
                 min_period=None, max_period=None, period=None, high_pass_cutoff=None):
        self.__dict__.update(locals())
        super(Repet, self).__init__(input_audio_signal=input_audio_signal,
                                    sample_rate=sample_rate, stft_params=stft_params)
        self.repet_type = RepetType.DEFAULT if repet_type is None else repet_type
        self.high_pass_cutoff = 100 if high_pass_cutoff is None else high_pass_cutoff

        if self.repet_type not in RepetType.all_types:
            raise TypeError('\'repet_type\' in Repet() constructor cannot be {}'.format(repet_type))

        if self.repet_type == RepetType.SIM:
            # if similarity_threshold == 0:
            #     warnings.warn('Using default value of 1 for similarity_threshold.')
            # if min_distance_between_frames == 1:
            #     warnings.warn('Using default value of 0 for min_distance_between_frames.')
            # if max_repeating_frames == 100:
            #     warnings.warn('Using default value of 100 for maxRepeating frames.')

            self.similarity_threshold = 0 if similarity_threshold is None else similarity_threshold
            self.min_distance_between_frames = 1 if min_distance_between_frames is None else min_distance_between_frames
            self.max_repeating_frames = 10 if max_repeating_frames is None else max_repeating_frames
        elif self.repet_type == RepetType.ORIGINAL:

            if period is None:
                self.min_period = 0.8 if min_period is None else min_period
                self.max_period = min(8, self.audio_signal.signal_duration / 3) if max_period is None else max_period
            else:
                self.period = period
                self.period = self._update_period(self.period)

        elif self.repet_type == RepetType.ADAPTIVE:
            raise NotImplementedError('Not allowed to do this yet!')

        self.verbose = False

    # @property
    def run(self):
        """Runs the REPET algorithm

        Returns:
            y (AudioSignal): repeating background (M by N) containing M channels and N time samples
            (the corresponding non-repeating foreground is equal to x-y)

        Example:
             ::
            signal = nussl.AudioSignal(path_to_input_file='input_name.wav')

            # Set up window parameters
            win = nussl.WindowAttributes(signal.sample_rate)
            win.window_length = 2048
            win.window_type = nussl.WindowType.HAMMING

            # Set up and run Repet
            repet = nussl.Repet(signal, window_type=nussl.RepetType.SIM, stft_params=win)
            repet.min_distance_between_frames = 0.1
            repet.run()

        """

        # unpack window parameters
        win_len, win_type, hop_len, nfft = self.stft_params.window_length, self.stft_params.window_type, \
                                           self.stft_params.hop_length, self.stft_params.n_fft_bins

        # High pass filter cutoff freq. (in # of freq. bins), +1 to match MATLAB implementation
        self.high_pass_cutoff = np.ceil(float(self.high_pass_cutoff) * (nfft - 1) / self.sample_rate) + 1

        self._compute_spectrum()

        # run the specific algorithm
        mask = None
        S = None
        if self.repet_type == RepetType.SIM:
            S = self._do_repet_sim()
            mask = self.compute_repeating_mask_sim

        elif self.repet_type == RepetType.ORIGINAL:
            S = self._do_repet_original()
            mask = self.compute_repeating_mask_with_beat_spectrum

        elif self.repet_type == RepetType.ADAPTIVE:
            raise NotImplementedError('How did you get into this state????')

        # separate the mixture background by masking
        N, M = self.audio_signal.audio_data.shape
        bkgd = np.zeros_like(self.audio_signal.audio_data)
        for i in range(N):
            repeating_mask = mask(self.magnitude_spectrogram[:, :, i], S)
            repeating_mask[1:self.high_pass_cutoff, :] = 1  # high-pass filter the foreground
            repeating_mask = np.vstack((repeating_mask, repeating_mask[-2:0:-1, :].conj()))
            stft_with_mask = repeating_mask * self.stft[:, :, i]
            y = spectral_utils.e_istft(stft_with_mask, win_len, hop_len, win_type,
                                       reconstruct_reflection=False, remove_padding=True)
            bkgd[i,] = y[0:M]

        self.bkgd = AudioSignal(audio_data_array=bkgd, sample_rate=self.sample_rate)

        return self.bkgd

    def _compute_spectrum(self):
        self.stft = self.audio_signal.stft(self.stft_params.window_length, self.stft_params.hop_length,
                                           self.stft_params.window_type, self.stft_params.n_fft_bins,
                                           overwrite=False, remove_reflection=False)

        self.magnitude_spectrogram = np.abs(self.stft[0:self.stft_params.window_length//2 + 1, :, :])

    def get_similarity_matrix(self):
        """Calculates and returns the similarity matrix for the audio file associated with this object

        Returns:
             similarity_matrix (np.array): similarity matrix for the audio file.

        EXAMPLE:
             ::
            # Set up audio signal
            signal = nussl.AudioSignal('path_to_file.wav')

            # Set up a Repet object
            repet = nussl.Repet(signal)

            # I don't have to run repet to get a similarity matrix for signal
            sim_mat = repet.get_similarity_matrix()

        """
        self._compute_spectrum()
        V = np.mean(self.magnitude_spectrogram, axis=2)
        self.similarity_matrix = self.compute_similarity_matrix(V)
        return self.similarity_matrix

    def get_beat_spectrum(self):
        """Calculates and returns the beat spectrum for the audio file associated with this object

        Returns:
            beat_spectrum (np.array): beat spectrum for the audio file

        EXAMPLE:
             ::
            # Set up audio signal
            signal = nussl.AudioSignal('path_to_file.wav')

            # Set up a Repet object
            repet = nussl.Repet(signal)

            # I don't have to run repet to get a beat spectrum for signal
            beat_spec = repet.get_beat_spectrum()
        """
        self._compute_spectrum()
        self.beat_spectrum = self.compute_beat_spectrum(np.mean(np.square(self.magnitude_spectrogram ** 2), axis=2))
        return self.beat_spectrum

    def _do_repet_sim(self):
        # unpack window overlap
        ovp = self.stft_params.window_overlap_samples

        Vavg = np.mean(self.magnitude_spectrogram, axis=2)
        S = self.compute_similarity_matrix(Vavg)

        self.min_distance_between_frames = np.round([self.min_distance_between_frames * self.sample_rate / ovp])
        S = self.find_similarity_indices(S)

        return S

    def _do_repet_original(self):
        self.min_period = self._update_period(self.min_period)
        self.max_period = self._update_period(self.max_period)
        self.beat_spectrum = self.compute_beat_spectrum(np.mean(np.square(self.magnitude_spectrogram), axis=2).T)
        self.repeating_period = self.find_repeating_period_simple(self.beat_spectrum, self.min_period, self.max_period)
        return self.repeating_period

    @staticmethod
    def compute_similarity_matrix(X):
        """Computes the similarity matrix using the cosine similarity for any given input matrix X.
        
        Parameters:
            X (np.array): 2D matrix containing the magnitude spectrogram of the audio signal (Lf by Lt)
        Returns:
            S (np.array): similarity matrix (Lt by Lt)


        """
        assert (type(X) == np.ndarray)

        # normalize the columns of the magnitude spectrogram
        Lt = X.shape[1]
        X = X.T
        for i in range(0, Lt):
            Xi = X[i, :]
            rowNorm = np.sqrt(np.dot(Xi, Xi))
            X[i, :] = Xi / (rowNorm + constants.EPSILON)

        # compute the similarity matrix    
        S = np.dot(X, X.T)
        return S

    def find_similarity_indices(self, S):
        """Finds the similarity indices for all time frames from the similarity matrix
        
        Parameters:
            S (np.array): similarity matrix (Lt by Lt)
                 
        Returns:
            I (np.array): similarity indices for all time frames
        """

        Lt = S.shape[0]
        I = np.zeros((Lt, self.max_repeating_frames))

        for i in range(0, Lt):
            pind = self.find_peaks(S[i, :], self.similarity_threshold,
                                   self.min_distance_between_frames, self.max_repeating_frames)
            I[i, :] = pind

        return I

    def find_peaks(self, data, min_thr=0.5, min_dist=None, max_num=1):
        """Finds the peak values and corresponding indices in a vector of real values in [0,1]
        
        Parameters:
            data (np.array): row vector of real values (in [0,1])
            min_thr: (Optional[float]) minimum threshold (in [0,1]) on data values. Defaults to 0.5
            min_dist:(Optional[int]) minimum distance (in # of elements) between peaks. Defaults to .25 * data.length
            max_num: (Optional[int]) maximum number of peaks. Defaults to 1
        
        Returns:
            peak_indices (np.array): sorted array of indices of peaks in the data
        """

        # make sure data is a Numpy matrix
        data = np.mat(data)

        lenData = data.shape[1]
        if min_dist is None:
            min_dist = np.floor(lenData / 4)

        peak_indices = np.zeros((1, max_num), int)

        data = np.multiply(data, (data >= min_thr))
        if np.size(np.nonzero(data)) < max_num:
            raise ValueError('not enough number of peaks! change parameters.')
        else:
            i = 0
            while i < max_num:
                peak_indices[0, i] = np.argmax(data)
                data[0, peak_indices[0, i] - min_dist - 1:peak_indices[0, i] + min_dist + 1] = 0
                i += 1
                if np.sum(data) == 0:
                    break

        peak_indices = np.sort(peak_indices)

        return peak_indices

    @staticmethod
    def compute_repeating_mask_sim(V, I):
        """Computes the soft mask for the repeating part using the magnitude spectrogram and the similarity indices

        Parameters:
            V (np.array): 2D matrix containing the magnitude spectrogram of a signal (Lf by Lt)
            I (np.array): array containing similarity indices for all time frames
        Returns:
            M (np.array): 2D matrix (Lf by Lt) containing the soft mask for the repeating part. Elements of M take on
            values in [0,1]
         """

        Lf, Lt = np.shape(V)
        W = np.zeros_like(V).T
        for i in range(0, Lt):
            pind = I[i, :]
            W[i, :] = np.median(V.T[pind.astype(int), :], axis=0)

        W = W.T
        Wrow = np.reshape(W, (1, Lf * Lt))
        Vrow = np.reshape(V, (1, Lf * Lt))
        W = np.min(np.vstack([Wrow, Vrow]), axis=0)
        W = np.reshape(W, (Lf, Lt))
        M = (W + constants.EPSILON) / (V + constants.EPSILON)

        return M

    @staticmethod
    def compute_beat_spectrum(X):
        """Computes the beat spectrum averages (over freq.s) the autocorrelation matrix of a one-sided spectrogram.

         The autocorrelation matrix is computed by taking the autocorrelation of each row of the spectrogram and
         dismissing the symmetric half.

        Parameters:
            X (np.array): 2D matrix containing the one-sided power
            spectrogram of the audio signal (Lf by Lt by num channels)
        Returns:
            b (np.array): array containing the beat spectrum based on the power spectrogram
        """
        # compute the row-wise autocorrelation of the input spectrogram
        Lf, Lt = X.shape
        X = np.vstack([X, np.zeros_like(X)])
        Fx = scifft.fft(X, axis=0)
        Ax = np.abs(Fx)
        Sx = Ax ** 2  # fft over columns (take the fft of each row at a time)
        Rx = np.real(scifft.ifft(Sx, axis=0)[0:Lf,:])  # ifft over columns
        NormFactor = np.tile(np.arange(Lf, 0, -1), (Lt, 1)).T  # normalization factor
        Rx = Rx / NormFactor

        # compute the beat spectrum
        b = np.mean(Rx, axis=1)  # average over frequencies

        return b

    @staticmethod
    def find_repeating_period_simple(beat_spectrum, min_period, max_period):
        """Computes the repeating period of the sound signal using the beat spectrum using simple algorithm.

        Parameters:
            beat_spectrum (np.array): input beat spectrum array
            min_period (int): minimum possible period value
            max_period (int): maximum possible period value
        Returns:
             period (int) : The period of the sound signal in stft time bins
        """

        beat_spectrum = beat_spectrum[1:]  # discard the first element of beat_spectrum (lag 0)
        beat_spectrum = beat_spectrum[min_period - 1:  max_period]
        period = np.argmax(beat_spectrum) + min_period

        return period

    @staticmethod
    def find_repeating_period_complex(beat_spectrum):
        auto_cosine = np.zeros((len(beat_spectrum), 1))

        for i in range(0, len(beat_spectrum) - 1):
            auto_cosine[i] = 1 - scipy.spatial.distance.cosine(beat_spectrum[0:len(beat_spectrum) - i],
                                                               beat_spectrum[i:len(beat_spectrum)])

        ac = auto_cosine[0:np.floor(auto_cosine.shape[0])/2]
        auto_cosine = np.vstack([ac[1], ac, ac[-2]])
        auto_cosine_diff = np.ediff1d(auto_cosine)
        sign_changes = auto_cosine_diff[0:-1]*auto_cosine_diff[1:]
        sign_changes = np.where(sign_changes < 0)[0]

        extrema_values = ac[sign_changes]

        e1 = np.insert(extrema_values, 0, extrema_values[0])
        e2 = np.insert(extrema_values, -1, extrema_values[-1])

        extrema_neighbors = np.stack((e1[0:-1], e2[1:]))

        m = np.amax(extrema_neighbors, axis=0)
        extrema_values = extrema_values.flatten()
        maxima = np.where(extrema_values >= m)[0]
        maxima = zip(sign_changes[maxima], extrema_values[maxima])
        maxima = maxima[1:]
        maxima = sorted(maxima, key = lambda x: -x[1])
        period = maxima[0][0]

        return period

    @staticmethod
    def compute_repeating_mask_with_beat_spectrum(V, p):
        """Computes the soft mask for the repeating part using the magnitude spectrogram and the repeating period

        Parameters:
            V (np.array): 2D matrix containing the magnitude spectrogram of a signal (Lf by Lt)
            p (int): repeating period measured in # of time frames
        Returns:
            M (np.array): 2D matrix (Lf by Lt) containing the soft mask for the repeating part, elements of M take on
            values in [0,1]

        """
        p += 1 # this is a kluge to make this implementation match the original MATLAB implementation
        n, m = V.shape
        r = np.ceil(float(m) / p)
        W = np.hstack([V, float('nan') * np.zeros((n, r * p - m))])
        W = np.reshape(W.T, (r, n * p))
        W1 = np.median(W[0:r, 0:n * (m - (r - 1) * p)], axis=0)
        W2 = np.median(W[0:r - 1, n * (m - (r - 1) * p):n * p], axis=0)
        W = np.hstack([W1, W2])
        W = np.reshape(np.tile(W, (r, 1)), (r * p, n)).T
        W = W[:, 0:m]

        Wrow = W.flatten()  # np.reshape(W, (1, Lf * Lt))
        Vrow = V.flatten()  # np.reshape(V, (1, Lf * Lt))
        W = np.min(np.vstack([Wrow, Vrow]), axis=0)
        W = np.reshape(W, (n, m))
        M = (W + constants.EPSILON) / (V + constants.EPSILON)

        return M

    def update_periods(self):
        """
        Will update periods for use with self.find_repeating_period_simple(). Updates from seconds to stft bin values.
        Call this if you haven't done self.run() or else you won't get good results
        Examples:
            ::
            a = nussl.AudioSignal('path/to/file.wav')
            r = nussl.Repet(a)

            beat_spectrum = r.get_beat_spectrum()
            r.update_periods()
            repeating_period = r.find_repeating_period_simple(beat_spectrum, r.min_period, r.max_period)

        """
        self.period = self._update_period(self.period) if self.period is not None else None
        self.min_period = self._update_period(self.min_period) if self.min_period is not None else None
        self.max_period = self._update_period(self.max_period) if self.max_period is not None else None

    def _update_period(self, period):
        period = float(period)
        result = period * self.audio_signal.sample_rate
        result += self.stft_params.window_length / self.stft_params.window_overlap - 1
        result /= self.stft_params.window_overlap
        return np.ceil(result)

    def plot(self, output_file, **kwargs):
        """ Creates a plot of either the beat spectrum or similarity matrix for this file and outputs to output_file.
            By default, the repet_type is used to determine which to plot,
            (original -> beat spectrum. sim -> similarity matrix)
            You can override this by passing in plot_beat_spectrum=True or plot_sim_matrix=True as parameters.
            You cannot set both of these overrides simultaneously.

        Parameters:
            output_file (string) : string representing a path to the desired output file to be created.
            plot_beat_spectrum (Optional[bool]) : Setting this will force plotting the beat spectrum
            plot_sim_matrix (Optional[bool]) : Setting this will force plotting the similarity matrix

        EXAMPLE:
             ::
            # To plot the beat spectrum you have a few options:

            # 1) (recommended)
            # set up AudioSignal
            signal = nussl.AudioSignal('path_to_file.wav')

            # by default, this Repet object is now set to the original repet (RepetType.ORIGINAL)
            repet1 = nussl.Repet(signal)

            # plots beat spectrum by default
            repet1.plot('new_beat_spec_plot.png')

            # 2)
            # by giving this Repet object RepetType.SIM, it will default to plotting the similarity matrix
            repet2 = nussl.Repet(signal, repet_type=nussl.RepetType.SIM)

            # but we can override this Repet object plotting the similarity matrix with this argument
            repet2.plot('new_sim_matrix_plot.png', plot_beat_spectrum=True)

            # To plot the similarity matrix you have a few options:
            # 3) (recommended)
            # set up AudioSignal
            signal = nussl.AudioSignal('path_to_file.wav')

            # by giving this Repet object RepetType.SIM, it will default to printing the similarity matrix
            repet3 = nussl.Repet(signal, repet_type=nussl.RepetType.SIM)

            # plots similarity matrix by default
            repet3.plot('new_sim_matrix_plot.png')

            # 4)
            # by default, this Repet object is now set to the original repet (RepetType.ORIGINAL)
            repet4 = nussl.Repet(signal)

            # BUT we can override plotting the beat spectrum with this argument
            repet4.plot('new_sim_matrix_plot.png', plot_sim_matrix=True)

            # NOTE: You cannot do
            # repet.plot('new_plot.png', plot_beat_spectrum=True, plot_sim_matrix=True)
            # this will cause nussl to throw an exception!
        """
        import matplotlib.pyplot as plt
        plt.close('all')

        plot_beat_spectrum = self.repet_type is RepetType.ORIGINAL
        plot_sim_matrix = self.repet_type is RepetType.SIM

        if kwargs is not None:
            if kwargs.has_key('plot_beat_spectrum'):
                plot_beat_spectrum = kwargs['plot_beat_spectrum']
            if kwargs.has_key('plt_sim_matrix'):
                plot_sim_matrix = kwargs['plot_sim_matrix']

        if plot_beat_spectrum == plot_sim_matrix == True:
            raise AssertionError('Cannot set both plot_beat_spectrum=True and plot_sim_matrix=True!')

        if plot_beat_spectrum:
            plt.plot(self.get_beat_spectrum())
            plt.title('Beat Spectrum for {}'.format(self.audio_signal.file_name))
            plt.grid('on')

        elif plot_sim_matrix:
            plt.pcolormesh(self.get_similarity_matrix())
            plt.title('Similarity Matrix for {}'.format(self.audio_signal.file_name))

        plt.axis('tight')
        plt.savefig(output_file)

    def make_audio_signals(self):
        """ Returns the background and foreground audio signals

        Returns:
            Audio Signals (List): 2 element list.

                * bkgd: Audio signal with the calculated background track
                * fkgd: Audio signal with the calculated foreground track

        EXAMPLE:
             ::
            # set up AudioSignal object
            signal = nussl.AudioSignal('path_to_file.wav')

            # set up and run repet
            repet = nussl.Repet(signal)
            repet.run()

            # get audio signals (AudioSignal objects)
            background, foreground = repet.make_audio_signals()
        """
        self.fgnd = self.audio_signal - self.bkgd
        return [self.bkgd, self.fgnd]


class RepetType():
    """Types of Repet algorithm implementation.
    """
    ORIGINAL = 'original'
    SIM = 'sim'
    # ADAPTIVE = 'adaptive'
    DEFAULT = ORIGINAL
    all_types = [ORIGINAL, SIM]

    def __init__(self):
        pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import division
import os.path
import numpy as np
import scipy.io.wavfile as wav
import librosa
import numbers
import audioread

import spectral_utils
import constants


class AudioSignal(object):
    """Defines properties of an audio signal and performs basic operations such as Wav loading and STFT/iSTFT.

    Parameters:
        path_to_input_file (string): string specifying path to file. Either this or timeSeries must be provided
        audio_data_array (np.array): Numpy matrix containing a time series of a signal
        signalStartingPosition (Optional[int]): Starting point of the section to be extracted in seconds. Defaults to 0
        signalLength (Optional[int]): Length of the signal to be extracted. Defaults to full length of the signal
        SampleRate (Optional[int]): sampling rate to read audio file at. Defaults to Constants.DEFAULT_SAMPLE_RATE
        stft (Optional[np.array]): Optional pre-coumputed complex spectrogram data.

    Attributes:
        window_type(WindowType): type of window to use in operations on the signal. Defaults to WindowType.DEFAULT
        window_length (int): Length of window in ms. Defaults to 0.06 * SampleRate
        num_fft_bins (int): Number of bins for fft. Defaults to windowLength
        overlap_ratio (float): Ratio of window that overlaps in [0,1). Defaults to 0.5
        stft_data (np.array): complex spectrogram of the data
        power_spectrum_data (np.array): power spectrogram of the data
        Fvec (np.array): frequency vector for stft
        Tvec (np.array): time vector for stft
        sample_rate (int): sampling frequency
  
    Examples:
        * create a new signal object:     ``sig=AudioSignal('sample_audio_file.wav')``
        * compute the spectrogram of the new signal object:   ``sigSpec,sigPow,F,T=sig.STFT()``
        * compute the inverse stft of a spectrogram:          ``sigrec,tvec=sig.iSTFT()``
  
    """

    def __init__(self, path_to_input_file=None, audio_data_array=None, signal_starting_position=0, signal_length=0,
                 sample_rate=constants.DEFAULT_SAMPLE_RATE, stft=None, stft_params=None):

        self.path_to_input_file = path_to_input_file
        self._audio_data = None
        self.sample_rate = sample_rate

        if (path_to_input_file is not None) and (audio_data_array is not None):
            raise Exception('Cannot initialize AudioSignal object with a path AND an array!')

        if path_to_input_file is not None:
            self.load_audio_from_file(self.path_to_input_file, signal_length, signal_starting_position)
        elif audio_data_array is not None:
            self.load_audio_from_array(audio_data_array, sample_rate)

        # stft data
        self.stft_data = np.array([]) if stft is None else stft  # complex spectrogram
        self.power_spectrum_data = np.array([])  # power spectrogram
        self.freq_vec = np.array([])  # freq. vector
        self.time_vec = np.array([])  # time vector

        self.stft_params = spectral_utils.StftParams(self.sample_rate) if stft_params is None else stft_params

    def __str__(self):
        return 'AudioSignal'

    ##################################################
    # Plotting
    ##################################################

    def plot_time_domain(self):
        """
        Not implemented yet -- will raise and exception
        Returns:

        """
        raise NotImplementedError('Not ready yet!')

    _NAME_STEM = 'audio_signal'

    def plot_spectrogram(self, file_name=None, ch=None):
        # TODO: make other parameters adjustable
        if file_name is None:
            name_stem = self.file_name if self.file_name is not None else self._NAME_STEM
        else:
            if os.path.isfile(file_name):
                name_stem = os.path.splitext(file_name)[0]
            else:
                name_stem = file_name

        if ch is None:
            if self.num_channels > 1:
                for i in range(1, self.num_channels+1):
                    name = name_stem + '_ch{}.png'.format(i)
                    spectral_utils.plot_stft(self.get_channel(i), name,
                                             sample_rate=self.sample_rate)
            else:
                name = name_stem + '.png'
                spectral_utils.plot_stft(self.get_channel(1), name,
                                         sample_rate=self.sample_rate)
        else:
            name = name_stem + '_ch{}.png'.format(ch)
            spectral_utils.plot_stft(self.get_channel(ch), name,
                                     sample_rate=self.sample_rate)

    ##################################################
    # Properties
    ##################################################

    # Constants for accessing _audio_data np.array indices
    _LEN = 1
    _CHAN = 0

    _STFT_LEN = 1
    _STFT_BINS = 0
    _STFT_CHAN = 2

    @property
    def signal_length(self):
        """Returns the length of the audio signal represented by this object in samples
        """
        return self._audio_data.shape[self._LEN]

    @property
    def signal_duration(self):
        """Returns the length of the audio signal represented by this object in seconds
        """
        return self.signal_length / self.sample_rate

    @property
    def num_channels(self):
        """The number of channels
        """
        return self._audio_data.shape[self._CHAN]

    @property
    def audio_data(self):
        """A numpy array that represents the audio
        """
        return self._audio_data

    @audio_data.setter
    def audio_data(self, value):
        assert (type(value) == np.ndarray)

        self._audio_data = value

        if self._audio_data.ndim < 2:
            self._audio_data = np.expand_dims(self._audio_data, axis=self._CHAN)

    @property
    def file_name(self):
        """The name of the file wth extension, NOT the full path
        """
        if self.path_to_input_file is not None:
            return os.path.split(self.path_to_input_file)[1]
        return None

    @property
    def time_vector(self):
        return np.array((1. / self.sample_rate) * np.arange(self.signal_length))

    ##################################################
    # I/O
    ##################################################

    def load_audio_from_file(self, input_file_path, signal_starting_position=0, signal_length=0):
        """Loads an audio signal from a .wav file

        Parameters:
            input_file_path: path to input file.
            signal_length (Optional[int]): Length of signal to load. signal_length of 0 means read the whole file
             Defaults to the full length of the signal
            signal_starting_position (Optional[int]): The starting point of the section to be extracted (seconds).
             Defaults to 0 seconds

        """

        self.path_to_input_file = input_file_path
        try:
            with audioread.audio_open(os.path.realpath(input_file_path)) as input_file:
                self.sample_rate = input_file.samplerate
                file_length = input_file.duration
                n_ch = input_file.channels

            if signal_length == 0:
                signal_length = file_length

            read_mono = True
            if n_ch != 1:
                read_mono = False

            audio_input, self.sample_rate = librosa.load(input_file_path,
                                                         sr=input_file.samplerate,
                                                         offset=signal_starting_position,
                                                         duration=signal_length,
                                                         mono=read_mono)

            # Change from fixed point to floating point
            if not np.issubdtype(audio_input.dtype, float):
                audio_input = audio_input.astype('float') / (np.iinfo(audio_input.dtype).max + 1.0)

            self.audio_data = audio_input

        except Exception, e:
            print "Cannot read from file, {file}".format(file=input_file_path)
            print "If you are convinced that this audio file should work, please use ffmpeg to reformat it."
            raise e

    def load_audio_from_array(self, signal, sample_rate=constants.DEFAULT_SAMPLE_RATE):
        """Loads an audio signal from a numpy array. Only accepts float arrays and int arrays of depth 16-bits.

        Parameters:
            signal (np.array): np.array containing the audio file signal sampled at sampleRate
            sample_rate (Optional[int]): the sample rate of signal. Default is Constants.DEFAULT_SAMPLE_RATE (44.1kHz)

        """
        assert (type(signal) == np.ndarray)

        self.path_to_input_file = None

        # Change from fixed point to floating point
        if not np.issubdtype(signal.dtype, float):
            if np.max(signal) > np.iinfo(np.dtype('int16')).max:
                raise ValueError('Please convert your array to 16-bit audio.')

            signal = signal.astype('float') / (np.iinfo(np.dtype('int16')).max + 1.0)

        self.audio_data = signal
        self.sample_rate = sample_rate

    def write_audio_to_file(self, output_file_path, sample_rate=None, verbose=False):
        """Outputs the audio signal to a .wav file

        Parameters:
            output_file_path (str): Filename where waveform will be saved
            sample_rate (Optional[int]): The sample rate to write the file at. Default is AudioSignal.SampleRate, which
            is the samplerate of the original signal.
            verbose (Optional[bool]): Flag controlling printing when writing the file.
        """
        if self.audio_data is None:
            raise Exception("Cannot write audio file because there is no audio data.")

        try:
            self.peak_normalize()

            if sample_rate is None:
                sample_rate = self.sample_rate

            audio_output = np.copy(self.audio_data)

            # TODO: better fix
            # convert to fixed point again
            if not np.issubdtype(audio_output.dtype, int):
                audio_output = np.multiply(audio_output, 2 ** (constants.DEFAULT_BIT_DEPTH - 1)).astype('int16')

            wav.write(output_file_path, sample_rate, audio_output.T)
        except Exception, e:
            print "Cannot write to file, {file}.".format(file=output_file_path)
            raise e
        if verbose:
            print "Successfully wrote {file}.".format(file=output_file_path)

    ##################################################
    #               STFT Utilities
    ##################################################

    def stft(self, window_length=None, hop_length=None, window_type=None, n_fft_bins=None, remove_reflection=True,
             overwrite=True):
        """Computes the Short Time Fourier Transform (STFT) of the audio signal

        Warning:
            If overwrite=True (default) this will overwrite any data in self.stft_data!

        Returns:
            * **stfts** (*np.array*) - complex stft data
        """
        if self.audio_data is None:
            raise Exception("No self.audio_data (time domain) to make STFT from!")

        window_length = self.stft_params.window_length if window_length is None else window_length
        hop_length = self.stft_params.hop_length if hop_length is None else hop_length
        window_type = self.stft_params.window_type if window_type is None else window_type
        n_fft_bins = self.stft_params.n_fft_bins if n_fft_bins is None else n_fft_bins

        calculated_stft = self._do_stft(window_length, hop_length, window_type, n_fft_bins, remove_reflection)

        if overwrite:
            self.stft_data = calculated_stft
        self.power_spectrum_data = np.array([])

        return calculated_stft

    def _do_stft(self, window_length, hop_length, window_type, n_fft_bins, remove_reflection):
        if self.audio_data is None:
            raise Exception('Cannot do stft without signal!')

        stfts = []

        for i in range(1, self.num_channels + 1):
            stfts.append(spectral_utils.e_stft(self.get_channel(i), window_length,
                                               hop_length, window_type, n_fft_bins, remove_reflection))

        return np.array(stfts).transpose((1, 2, 0))

    def istft(self, window_length=None, hop_length=None, window_type=None, n_fft_bins=None, overwrite=True,
              reconstruct_reflection=True):
        """Computes and returns the inverse Short Time Fourier Transform (STFT).

        Warning:
            If overwrite=True (default) this will overwrite any data in self.audio_data!

        Returns:
             * **calculated_signal** (np.array): time-domain audio signal
        """
        if self.stft_data.size == 0:
            raise Exception('Cannot do inverse STFT without self.stft_data!')

        window_length = self.stft_params.window_length if window_length is None else window_length
        hop_length = self.stft_params.hop_length if hop_length is None else hop_length
        window_type = self.stft_params.window_type if window_type is None else window_type
        n_fft_bins = self.stft_params.n_fft_bins if n_fft_bins is None else n_fft_bins

        calculated_signal = self._do_istft(window_length, hop_length, window_type, n_fft_bins, reconstruct_reflection)

        if overwrite:
            self.audio_data = calculated_signal

        return calculated_signal

    def _do_istft(self, window_length, hop_length, window_type, n_fft_bins, reconstruct_reflection):
        if self.stft_data.size == 0:
            raise ('Cannot do inverse STFT without self.stft_data!')

        signals = []
        for i in range(self.num_channels):
            signals.append(
                spectral_utils.e_istft(self.get_stft_channel(i + 1), window_length, hop_length, window_type,
                                       reconstruct_reflection))

        return np.array(signals)

    ##################################################
    #                  Utilities
    ##################################################

    def concat(self, other):
        """ Add two AudioSignal objects (by adding self.audio_data) temporally.

        Parameters:
            other (AudioSignal): Audio Signal to concatenate with the current one.
        """
        if self.num_channels != other.num_channels:
            raise Exception('Cannot concat two signals that have a different number of channels!')

        if self.sample_rate != other.sample_rate:
            raise Exception('Cannot add two signals that have different sample rates!')

        self.audio_data = np.concatenate((self.audio_data, other.audio_data), axis=self._LEN)

    def truncate_samples(self, n_samples):
        """ Truncates the signal leaving only the first n_samples number of samples.
        """
        if n_samples > self.signal_length:
            raise Exception('n_samples must be less than self.signal_length!')

        self.audio_data = self.audio_data[:, 0: n_samples]

    def trancate_seconds(self, n_seconds):
        """ Truncates the signal leaving only the first n_seconds
        """
        if n_seconds > self.signal_duration:
            raise Exception('n_seconds must be shorter than self.signal_duration!')

        n_samples = n_seconds * self.sample_rate
        self.truncate_samples(n_samples)

    def zero_pad(self, before, after):
        """
        Adds zeros before and after the signal to all channels. Extends the length
        of self.audio_data by before + after.
        Args:
            before: (int) number of zeros to be put before the current contents of self.audio_data
            after: (int) number of zeros to be put after the current contents fo self.audio_data

        """
        for ch in range(1, self.num_channels + 1):
            self.audio_data = np.lib.pad(self.get_channel(ch), (before, after), 'constant', constant_values=(0, 0))

    def get_channel(self, n):
        """Gets the n-th channel from ``self.audio_data``. **1-based.**

        Parameters:
            n (int): index of channel to get. **1-based**
        Returns:
            n-th channel (np.array): the audio data in the n-th channel of the signal
        """
        if n > self.num_channels:
            raise Exception(
                'Cannot get channel {1} when this object only has {2} channels!'.format((n, self.num_channels)))

        if n <= 0:
            raise Exception('Cannot get channel {}. This will cause unexpected results'.format(n))

        return self.audio_data[n - 1, ]


    def get_stft_channel(self, n):
        """
        Returns the n-th channel from ``self.stft_data``. **1-based.**
        Args:
            n: (int) index of stft channel to get. **1 based**

        Returns:
            n-th channel (np.array): the stft data in the n-th channel of the signal
        """
        if n > self.num_channels:
            raise Exception(
                'Cannot get channel {1} when this object only has {2} channels!'.format((n, self.num_channels)))

        return self.stft_data[:, :, n - 1]

    def peak_normalize(self):
        """ Normalizes the whole audio file to 1.0.
            Notes:
            **If self.audio_data is not represented as floats this will convert the representation to floats!**

        """
        max_val = 1.0
        max_signal = np.max(np.abs(self.audio_data))
        if max_signal > max_val:
            self.audio_data = self.audio_data.astype('float') / max_signal

    def add(self, other):
        """Adds two audio signal objects. This does element-wise addition on the ``self.audio_data`` array.

        Parameters:
            other (AudioSignal): Other audio signal to add.
        Returns:
            sum (AudioSignal): AudioSignal with the sum of the current object and other.
        """
        return self + other

    def sub(self, other):
        """Subtracts two audio signal objects. This does element-wise subtraction on the ``self.audio_data`` array.

        Parameters:
            other (AudioSignal): Other audio signal to subtract.
        Returns:
            diff (AudioSignal): AudioSignal with the difference of the current object and other.
        """
        return self - other

    ##################################################
    #              Operator overloading
    ##################################################

    def __add__(self, other):
        self._verify_audio(other)

        # for ch in range(self.num_channels):
        # TODO: make this work for multiple channels
        if self.audio_data.size > other.audio_data.size:
            combined = np.copy(self.audio_data)
            combined[0: other.audio_data.size] += other.audio_data
        else:
            combined = np.copy(other.audio_data)
            combined[0: self.audio_data.size] += self.audio_data

        return AudioSignal(audio_data_array=combined)

    def __sub__(self, other):
        self._verify_audio(other)

        # for ch in range(self.num_channels):
        # TODO: make this work for multiple channels
        if self.audio_data.size > other.audio_data.size:
            combined = np.copy(self.audio_data)
            combined[0: other.audio_data.size] -= other.audio_data
        else:
            combined = np.copy(other.audio_data)
            combined[0: self.audio_data.size] -= self.audio_data

        return AudioSignal(audio_data_array=combined)

    def _verify_audio(self, other):
        if self.num_channels != other.num_channels:
            raise Exception('Cannot subtract two signals that have a different number of channels!')

        if self.sample_rate != other.sample_rate:
            raise Exception('Cannot subtract two signals that have different sample rates!')

    def __iadd__(self, other):
        return self + other

    def __isub__(self, other):
        return self - other

    def __mul__(self, other):
        assert isinstance(other, numbers.Real)
        raise NotImplemented('Not implemented yet.')

    def __len__(self):
        return self.signal_length



import os
import nussl


def main():
    # input audio file
    input_name = os.path.join('..', 'Input','mix3.wav')
    signal = nussl.AudioSignal(path_to_input_file=input_name)

    # make a directory to store output if needed
    if not os.path.exists('../Output/'):
        os.mkdir('../Output')

    # Set up Repet
    repet = nussl.Repet(signal)
    L = 2048
    repet.stft_params.window_length = L
    repet.stft_params.window_type = nussl.WindowType.HAMMING
    repet.stft_params.hop_length = L // 2
    repet.stft_params.n_fft_bins = L
    # and Run
    repet.run()

    # Get foreground and backgroun audio signals
    bkgd, fgnd = repet.make_audio_signals()

    # and write out to files
    bkgd.write_audio_to_file(os.path.join('..', 'Output', 'repet_background.wav'))
    fgnd.write_audio_to_file(os.path.join('..', 'Output', 'repet_foreground.wav'))

    # Set up Repet
    # repet = nussl.Repet(signal, repet_type=nussl.RepetType.SIM)
    # # and Run
    # # repet.run()
    #
    # # Get foreground and backgroun audio signals
    # bkgd, fgnd = repet.make_audio_signals()
    #
    # # and write out to files
    # bkgd.write_audio_to_file(os.path.join('..', 'Output', 'repet_background_SIM.wav'))
    # fgnd.write_audio_to_file(os.path.join('..', 'Output', 'repet_foreground_SIM.wav'))


if __name__ == '__main__':
    main()



import random
import time
import numpy as np
import os

import nussl


def audio_example():
    """
    A simple source separation with audio files. Inputs two files of piano notes
    and adds them together and then creates two "guesses" for NMF to start from.
    Outputs files
    :return:
    """

    if not os.path.exists(os.path.join('..', 'Output/')):
        os.mkdir(os.path.join('..', 'Output/'))

    numNotes = 2

    # Two input files
    firstFileName = os.path.join('..', 'Input', 'K0140.wav')
    secondFileName = os.path.join('..', 'Input', 'K0149.wav')

    firstNote = nussl.AudioSignal(firstFileName)
    secondNote = nussl.AudioSignal(secondFileName)

    # Combine notes into one file and save target
    bothNotesPre = firstNote + secondNote
    bothNotesPre.write_audio_to_file(os.path.join('..', 'Output', 'combined_preNMF.wav'))

    bothNotesForNMF = firstNote + secondNote
    _, stft, _, _ = bothNotesForNMF.stft()

    # Make some 'guesses'
    jitter = 0.2
    max = 2 ** 15
    for i, val in np.ndenumerate(firstNote.audio_data):
        firstNote.audio_data[i] += int(float(random.random() * jitter) * max)
    _, firstStft, _, _ = firstNote.stft()
    firstGuessAct = np.sum(firstStft, axis=0)
    firstGuessVec = np.sum(firstStft, axis=1)

    for i, val in np.ndenumerate(secondNote.audio_data):
        secondNote.audio_data[i] += int(float(random.random() * jitter) * max)
    _, secondStft, _, _ = secondNote.stft()
    secondGuessAct = np.sum(secondStft, axis=0)
    secondGuessVec = np.sum(secondStft, axis=1)

    # put them into activation matrix and template vectors
    GuessVec = np.array([firstGuessVec, secondGuessVec]).T
    GuessAct = np.array([firstGuessAct, secondGuessAct])

    # run NMF
    nmf = nussl.Nmf(stft, numNotes)
    nmf.max_num_iterations = 100
    start = time.time()
    nmf.run()
    print '{0:.3f}'.format(time.time() - start), 'seconds for NUSSL'

    # Make output files
    outFileNameBase = os.path.join('..', 'Output', 'NMFoutput_')
    i = 1
    newSignals = nmf.make_audio_signals()
    for signal in newSignals:
        outFileName = outFileNameBase + str(i) + '.wav'
        signal.write_audio_to_file(outFileName)
        i += 1

    # Recombine signals and make a new output file
    recombined = newSignals[0] + newSignals[1]
    recombined.write_audio_to_file(os.path.join('..', 'Output', 'combined_postNMF.wav'))


if __name__ == "__main__":
    audio_example()



"""
Audio mixture generator implementation. The mixture types include: instantaneous,
anechoic, and convolutive. 

Required packages:
    * Numpy
    * Scipy
    * Matplotlib

"""

import numpy as np
from scipy.fftpack import fft, ifft
import matplotlib.pyplot as plt


def mkmixture(sn, mixparam, fs, rsch=False):
    """Synthesizes an M-channel convolutive mixture of N sources
    
    Parameters:
        sn (np.array): Numpy N by Ls array. Each row contains a single channel time-domain source signal.
        mixparam (List[np.array]): List containing locations of sources/mics and room parameters

            * Numpy N by 3 array containing source locations (x,y,z coordinates)
            * Numpy M by 3 array containing mic locations (x,y,z, coordinates)
            * Numpy 1 by 2 array containing room characteristics

                #. number of virtual sources
                #. reflection coefficients in (-1,1)

            * Numpy 1 by 3 array containing room dimensions (meters)

        fs (int): sampling rate (in Hz)
        rsch (Optional[bool]): logical argument indicating whether the room schematic should be generated or not.
         Default is False
          
    Returns:
        * **Mixtures** (*np.array*) - a Numpy M by Lx array. Each row contains a single channel time-domain mixture

        * **SourceIMS** (*np.array*) - a Numpy N by M by Lx array containing time-domain samples of source images

        * **H** (*list*) - a M by N list containing impulse responses of M*N acoustic channels between sources and mics
    """

    # find the number of sources and signal lengths
    N, Ls = np.shape(sn)

    # extract locations of sources and mics
    Ps = mixparam[0]
    Pm = mixparam[1]
    M = Pm.shape[0]

    # extract room parameters
    NumVir = mixparam[2][0]
    RefCoeff = mixparam[2][1]
    Rdim = mixparam[3]

    # compute room impulse response for source-mic pairs

    H = []
    Lh = np.zeros((M, N))
    for i in range(0, M):
        hm = []
        for j in range(0, N):
            hmn = rir(Rdim, Pm[i, :], Ps[j, :], NumVir, RefCoeff, fs)
            hm.append(hmn)
            Lh[i, j] = np.size(hmn)
        H.append(hm)

    # compute mixtures
    Lhmax = int(Lh.max())
    Lx = Ls + Lhmax - 1
    Mixtures = np.zeros((M, Lx))
    SourceIMS = np.zeros((N, M, Lx))

    for i in range(0, M):
        for j in range(0, N):
            hmn = H[i][j]
            xm = fastconv(np.array(sn[j, :], ndmin=2), hmn)
            Lxm = np.shape(xm)[1]
            SourceIMS[j, i, 0:Lxm] = xm
            Mixtures[i, 0:Lxm] = Mixtures[i, 0:Lxm] + xm

    # plot the recording setup if specified

    if rsch:
        SMarker = dict(color='red', linestyle='', marker='o',
                       markersize=10, markerfacecoloralt='red')
        plt.plot(Ps[:, 0], Ps[:, 1], **SMarker)

        MMarker = dict(color='blue', linestyle='', marker='s',
                       markersize=6, markerfacecoloralt='blue')

        plt.plot(Pm[:, 0], Pm[:, 1], **MMarker)
        plt.grid()
        plt.xlim((0, Rdim[0]))
        plt.ylim((0, Rdim[1]))
        plt.legend(['Sources', 'Mics'])

    return Mixtures, SourceIMS, H


def fastconv(x, h):
    """Performs fast convolution (in freq. domain rather than in time domain).
    
    Parameters:
        x (np.array): Numpy row vector containing samples of the time-domain signal x[n]
        h (np.array): Numpy row vector containing samples of the time-domain signal h[n] which here is considered to be
         the impulse response of a channel

    Returns:
        y (np.array): a Numpy row vector containing samples of the time-domain signal y[n], the output of the
        channel h to an input signal x

    References:
        The original MATLAB implementation can be found on the Mathworks File Exchange at:
        http://www.mathworks.com/matlabcentral/fileexchange/5110-fast-convolution
    """

    Ly = x.shape[1] + h.shape[1] - 1
    Ly2 = int(2 ** (np.ceil(np.log2(Ly))))

    X = fft(x, n=Ly2)
    H = fft(h, n=Ly2)
    Y = X * H
    y = np.real(ifft(Y, n=Ly2))
    y = np.array(y[0, 0:Ly], ndmin=2)

    y = y * (np.abs(x).max() / np.abs(y).max())

    return y


def rir(Rdim, Mcoords, Scoords, Numvs, Rcoef, fs):
    """Computes the room impulse response given the recording setup and the room features, using the mirror image method

    Parameters:
        Rdim (np.array): Numpy array containing the dimensions of the room (meters)
        Mcoords (np.array): Numpy 1 by 3 array containing microphone location (x,y,z, coordinates)
        Scoords (np.array): Numpy 1 by 3 array containing the sound source location (x,y,z, coordinates)
        Numvs (int): number of virtual sources will be (2*Numvs+1)**3
        Rcoef (int): reflection coefficinet for the walls, taking on a value in (-1,1)
        fs (int): sampling rate
    
    Returns:
        h (np.array): Numpy 1 by Lh array containing the room impulse response

    Example:
         ::
        Rdim=np.array([20,19,21])
        Mcoords=np.array([19,18,1.6])
        Scoords=np.array([5,2,1]
        Numvs=1
        Rcoef=0.3
        fs=44100
        h = rir(Rdim,Mcoords,Scoords,Numvs,Rcoef,fs)

    References:
        The original MATLAB implementation by Stephen G. McGovern can be found on the Mathworks File Exchange at:
        http://www.mathworks.com/matlabcentral/fileexchange/5116-room-impulse-response-generator

        McGovern, Stephen G. "Fast image method for impulse response calculations of box-shaped rooms."
        Applied Acoustics 70.1 (2009): 182-189.
    """

    Ind = np.arange(-Numvs, Numvs + 1)
    RMS = Ind + 0.5 - 0.5 * ((-1) ** Ind)
    SRCS = (-1) ** Ind

    xi = SRCS * Scoords[0] + RMS * Rdim[0] - Mcoords[0]
    yj = SRCS * Scoords[1] + RMS * Rdim[1] - Mcoords[1]
    zk = SRCS * Scoords[2] + RMS * Rdim[2] - Mcoords[2]

    i, j, k = np.meshgrid(xi, yj, zk)
    d = np.sqrt(i ** 2 + j ** 2 + k ** 2)
    time = (np.round(fs * d / 343) + 1).astype(int)

    e, f, g = np.meshgrid(Ind, Ind, Ind)
    c = Rcoef ** (np.abs(e) + np.abs(f) + np.abs(g))
    E = c / d

    time = np.reshape(time, (time.shape[0] * time.shape[1] * time.shape[2], 1))
    E = np.reshape(E, (E.shape[0] * E.shape[1] * E.shape[2], 1))

    h = np.zeros((1, time.max()))
    h[0, time[:, 0] - 1] = E.T

    return h



# In this demo the RPCA-based source separation method is tested

from scipy.io.wavfile import read,write
import matplotlib.pyplot as plt
plt.interactive('True')
import numpy as np
from f_stft import f_stft
from nussl.RPCA import rpca_ss
import time  

# close all figure windows
plt.close('all')

# load the audio file
#fs,x = read('/Users/fpishdadian/SourceSeparation/Audio Samples/Input/Sample2.wav')
fs,x = read('/Users/fpishdadian/SourceSeparation/Audio Samples/Input/mix5.wav')
x=np.array(x,ndmin=2)

# scale to -1.0 to 1.0
convert_16_bit = float(2**15)
x = x / (convert_16_bit + 1.0)

t=np.mat(np.arange(np.shape(x)[1])/float(fs))

# generate and plot the spectrogram of the mixture   
L=2048
win='Hamming'
ovp=0.75*L
nfft=L
mkplot=1
fmax=fs/2

plt.figure(1)
plt.title('Mixture')
Sm = f_stft(np.mat(x),L,win,ovp,fs,nfft,mkplot,fmax) 

# spactrogram parameters
specparam = [L,'Hamming',ovp,nfft]

# separation
start_time = time.clock()
y_f,y_b = rpca_ss(x,fs,specparam,mask=True,maskgain=1)
print time.clock() - start_time, "seconds"  


# normalize output signals
fg=y_f/np.abs(y_f).max()
bg=y_b/np.abs(y_b).max()

# plot the background and foreground
plt.figure(3)
plt.subplot(2,1,1)
plt.title('Background time-domain signal')
plt.plot(t.T,bg.T)
plt.axis('tight')
plt.show()
plt.subplot(2,1,2)
plt.title('Foreground time-domain signal')
plt.plot(t.T,fg.T)
plt.axis('tight')
plt.show()

plt.figure(4)
plt.subplot(2,1,1)
plt.title('Background Spectrogram')
Sb = f_stft(np.mat(y_b),L,win,ovp,fs,nfft,mkplot,fmax) 
plt.show()
plt.subplot(2,1,2)
plt.title('Foreground Spectrogram')
Sf = f_stft(np.mat(y_f),L,win,ovp,fs,nfft,mkplot,fmax) 
plt.show()

# check whether the separated spectrograms add up to the original spectrogram
Spec_diff=np.abs(Sm[0] - (Sb[0]+Sf[0]))

if Spec_diff.max()<1e-10:
    print('Background and foreground add up to the origianl mixture.')
    
# record the separated background and foreground in .wav files
filePath='/Users/fpishdadian/SourceSeparation/Audio Samples/Output/'
write(filePath+'rpcaBackground.wav',fs,bg.T)
write(filePath+'rpcaForeground.wav',fs,fg.T)


import time
import numpy as np
import matplotlib.pylab as plt
import os

import nussl


def sine_example():
    if not os.path.exists(os.path.join('..', 'Output/')):
        os.mkdir(os.path.join('..', 'Output/'))

    # Make signals
    nSamples = 44100  # 1 second per each frequency

    sin1 = np.sin(np.linspace(0, 100 * 2 * np.pi, nSamples))  # Freq = 100 Hz
    sin2 = np.sin(np.linspace(0, 200 * 2 * np.pi, nSamples))  # Freq = 200 Hz
    sin3 = np.sin(np.linspace(0, 300 * 2 * np.pi, nSamples))  # Freq = 300 Hz

    sines = np.concatenate((sin1, sin2, sin3))

    # load into AudioSignal object and get stft
    signal = nussl.AudioSignal(audio_data_array=sines)
    signal.stft()
    stft = signal.get_stft_channel(1)

    # Start NMF and time it
    start = time.time()
    nmf = nussl.Nmf(stft, 3)
    activation, dictionary = nmf.run()
    print '{0:.3f}'.format(time.time() - start), 'sec'

    # plot results
    plt.imshow(activation, interpolation='none', aspect='auto')
    ax = plt.axes()
    ya = ax.get_yaxis()
    ya.set_major_locator(plt.MaxNLocator(integer=True))
    xa = ax.get_xaxis()
    xa.set_major_locator(plt.MaxNLocator(integer=True))
    plt.title('Activation Matrix (H)')
    plt.savefig('../Output/A.png')
    plt.close()

    plt.imshow(dictionary, interpolation='none', aspect='auto')
    ax = plt.axes()
    ya = ax.get_yaxis()
    ya.set_major_locator(plt.MaxNLocator(integer=True))
    xa = ax.get_xaxis()
    xa.set_major_locator(plt.MaxNLocator(integer=True))
    plt.ylim([0, 20])
    plt.title('Template dictionary (W)')
    plt.savefig('../Output/B.png')


if __name__ == "__main__":
    sine_example()



# in this script a mixture of harmonic sources will be analyzed for separation uisng KAM
# along with harmonic kernels

import matplotlib.pyplot as plt
plt.interactive('True')
import numpy as np
from nussl.KAM import AudioSignal,kam
import time  

# close all figure windows
plt.close('all')

# load the audio mixture and generate spectrograms

FileName='/Users/fpishdadian/SourceSeparation/Audio Samples/Input/piano_mix2.wav'
mix=AudioSignal(FileName)

WinL=2*2048 # 93 ms window
Ovp=3*WinL/4 # 50% overlap   
mix.windowlength=WinL
mix.overlap_samples=Ovp
mix.num_fft_bins=WinL
mix.makeplot=1; 
mix.fmaxplot=5000;

plt.figure(1)
mix.do_STFT()
plt.title('Mixture')


# inputs of the 'kam' function

Inputfile=[FileName,'full length',0]

# define harmonic kernels with different fundamental freq.s
Np=1
SourceKernels=[]
SourceKernels.append(['harmonic',np.mat([24,Np])])  # source #1
SourceKernels.append(['harmonic',np.mat([32,Np])])  # source #2 
SourceKernels.append(['harmonic',np.mat([36,Np])])  # source #3 
SourceKernels.append(['harmonic',np.mat([40,Np])])  # source #4 

# define smooth harmonic kernels with differnt fundamental freq.s
#SourceKernels=[]
#P=np.array([24,32,36,40])
#NP=1
#Df=NP*P+1 
#def ismember(A, B):
#    return np.sum(np.array([ i == B.T for i in A.T ]),axis=1).T
#for i in range(0,4):    
#   Nhood=lambda TFcoords1,TFcoords2: np.logical_and(np.logical_and((np.tile(TFcoords1[:,1],(1,TFcoords2.shape[0]))==np.tile(TFcoords2[:,1].T,(TFcoords1.shape[0],1))),\
#                        (np.abs(np.tile(TFcoords1[:,0],(1,TFcoords2.shape[0]))-np.tile(TFcoords2[:,0].T,(TFcoords1.shape[0],1)))<Df[i])),\
#                        (ismember(np.mod(np.tile(TFcoords1[:,0],(1,TFcoords2.shape[0]))-np.tile(TFcoords2[:,0].T,(TFcoords1.shape[0],1)),P[i]),np.mat([0,1,P[i]-1])))) 
#   SourceKernels.append(['userdef',Nhood])
   

SpecParams=np.zeros(1,dtype=[('windowlength',int),('overlap_samples',int),('num_fft_bins',int)])
SpecParams['windowlength']=WinL
SpecParams['overlap_samples']=Ovp
SpecParams['num_fft_bins']=WinL

Numit=3

# call the kam function and record the running time
start_time = time.clock()
shat,fhat=kam(Inputfile,SourceKernels,Numit=Numit,SpecParams=SpecParams,FullKernel=False)[0:2]
print time.clock() - start_time, "seconds"   

# write the separated sources to .wav files
Ns=len(SourceKernels)
OutPath='/Users/fpishdadian/SourceSeparation/Audio Samples/Output/'
for i in range(0,Ns):
    ssi=AudioSignal(audiosig=shat[:,:,i],fs=mix.fs)  
    ssi.writeaudiofile(OutPath+'kamHout'+str(i+1)+'.wav')
    
 
 # plot the separated time-domain signals and corresponding power spectral dencities
ts=np.mat(np.arange(shat.shape[0])/float(mix.fs))
Fvec=mix.freq_vec
Tvec=mix.time_vec[0:fhat.shape[1]]
TT=np.tile(Tvec,(len(Fvec),1))
FF=np.tile(Fvec.T,(len(Tvec),1)).T


plt.figure(2)
for i in range(0,Ns):
    plt.subplot(Ns,1,i+1)
    #plt.plot(ts.T,src1.x[0:shat.shape[0]])
    plt.plot(ts.T,shat[:,0,i])
    plt.ylabel(r'$\hat{s}_'+str(i+1)+'(t)$')
    plt.axis('tight')
plt.xlabel('t(s)')    
    
plt.figure(3)
for i in range(0,Ns):    
    plt.subplot(Ns,1,i+1)
    plt.pcolormesh(TT,FF,np.log10(fhat[:,:,i]))
    plt.ylabel('f(Hz)')
    plt.title(r'$\hat{f}_'+str(i+1)+' $')
    plt.axis('tight')
    #plt.ylim(src1.freq_vec[0],5000)
plt.xlabel('t(s)')     
    



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
A repository containing all of the constants frequently used in
this source separation stuff.
"""

DEFAULT_SAMPLE_RATE = 44100
DEFAULT_WIN_LEN_PARAM = 0.04
DEFAULT_BIT_DEPTH = 16
DEFAULT_MAX_VAL = 2 ** 16
EPSILON = 1e-16
MAX_FREQUENCY = DEFAULT_SAMPLE_RATE // 2



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from .spectral_utils import *
from constants import *
from .audio_signal import AudioSignal
from separation_base import SeparationBase
from Duet import Duet
from Nmf import Nmf, DistanceType
from Repet import Repet, RepetType


__version__ = '0.1.5a3'

__title__ = 'nussl'
__description__ = 'A flexible sound source separation library.'
__uri__ = 'https://github.com/interactiveaudiolab/nussl'

__author__ = 'C. Grief, E. Manilow, F. Pishdadian'
__email__ = 'ethanmanilow2015@u.northwestern.edu'

__license__ = 'MIT'
__copyright__ = 'Copyright (c) 2016 Interactive Audio Lab'



# In this demo the KAM algorithm is tested

import matplotlib.pyplot as plt

plt.interactive('True')
import numpy as np
from nussl.KAM import AudioSignal, kaml
import time

# close all the figure windows
plt.close('all')

# Example (1): Test with a very simple scenario
#              The mixture includes a single frequency sinusoid (a horizontal line
#              in the TF domain) and two very short pulses (two vertical lines in
#              the TF domain)

# time signal
fs = 44100  # sampling rate
Ls = fs  # length of the signal (1 sec)
ts = np.mat(np.arange(Ls) / float(fs))
x1 = np.cos(2 * np.pi * 430.6640625 * ts)  # sinusoid
x2 = np.zeros((1, Ls))
x2[0, 11000:11050] = 10  # first pulse
x2[0, 22000:22050] = 10  # second pulse
x3 = (x1 + x2).T

# spectrogram
WL = 1024
sig = AudioSignal.AudioSignal(audioSignal=x3, fs=fs)
sig.makeplot = 1
sig.fmaxplot = fs / 2
sig.windowlength = WL
sig.nfft = WL
sig.overlapSamp = WL / 2

plt.figure(1)
plt.subplot(211)
plt.plot(ts.T, x3)
plt.xlabel('t (sec)')
plt.ylabel('x(t)')
plt.title('Mixture')
plt.subplot(212)
sig.do_STFT()
TT = np.tile(sig.Tvec, (len(sig.Fvec), 1))
FF = np.tile(sig.Fvec.T, (len(sig.Tvec), 1)).T

# inputs of the 'kam' function 
Inputfile = [x3, fs]
SourceKernels = [['horizontal', np.mat('10')], ['vertical', np.mat('10')]]  # 100ms horizontal and 387Hz vertical
SpecParams = np.zeros(1, dtype=[('windowlength', int), ('overlap_samples', int), ('num_fft_bins', int)])
SpecParams['windowlength'] = WL
SpecParams['overlap_samples'] = int(WL / 2)
SpecParams['num_fft_bins'] = WL
AlgParams = np.array([10, 1])
Numit = 2

# call the kam function and record the running time
start_time = time.clock()
shat, fhat = kaml(Inputfile, SourceKernels, AlgParams, Numit, SpecParams)[0:2]
print time.clock() - start_time, "seconds"

# record separated sources in .wav files
ss1 = AudioSignal(audiosig=shat[:, :, 0], fs=fs)
ss1.writeaudiofile('/Users/fpishdadian/SourceSeparation/Audio Samples/Output/kamOutSource1.wav')

ss2 = AudioSignal(audiosig=shat[:, :, 1], fs=fs)
ss2.writeaudiofile('/Users/fpishdadian/SourceSeparation/Audio Samples/Output/kamOutSource2.wav')

# plot the separated time-domain signals and corresponding power spectral densities
plt.figure(2)
plt.subplot(2, 2, 1)
plt.plot(ts.T, shat[:, :, 0])
plt.ylabel(r'$\hat{s}_1(t)$')
plt.xlabel('t(sec)')
plt.axis('tight')
plt.subplot(2, 2, 2)
plt.pcolormesh(TT, FF, np.log10(fhat[:, :, 0]))
plt.xlabel('t(sec)')
plt.ylabel('f(Hz)')
plt.title(r'$\hat{f}_1$')
plt.axis('tight')
plt.ylim(sig.Fvec[0], 5000)

plt.subplot(2, 2, 3)
plt.plot(ts.T, shat[:, 0, 1])
plt.ylabel(r'$\hat{s}_2(t)$')
plt.xlabel('t(sec)')
plt.axis('tight')
plt.subplot(2, 2, 4)
plt.pcolormesh(TT, FF, np.log10(fhat[:, :, 1]))
plt.xlabel('t(sec)')
plt.ylabel('f(Hz)')
plt.title(r'$\hat{f}_2$')
plt.axis('tight')
plt.ylim(sig.Fvec[0], 5000)



# Example (2): Test with a mixture of music and speech
#              The mixture is composed of flute, kick drum and female voice
# Note: the second examples takes much longer to run than the first example

# generate spectrograms of the separate sources 
WinL = 4096  # 93 ms window
Ovp = WinL / 2  # 50% overlap

src1 = AudioSignal('/Users/fpishdadian/SourceSeparation/Audio Samples/Input/src1.wav')
fs = src1.fs
src1.windowlength = WinL
src1.overlap_samples = Ovp
src1.num_fft_bins = WinL
src1.makeplot = 1
src1.fmaxplot = 1000

src2 = AudioSignal('/Users/fpishdadian/SourceSeparation/Audio Samples/Input/src2.wav')
src2.windowlength = WinL
src2.overlap_samples = Ovp
src2.num_fft_bins = WinL
src2.makeplot = 1
src2.fmaxplot = 2000

plt.figure(3)
plt.subplot(211)
src1.do_STFT()
plt.title('Drum')
plt.subplot(212)
src2.do_STFT()
plt.title('Flute')


# generate spectrograms of the mixture
mix1 = AudioSignal('/Users/fpishdadian/SourceSeparation/Audio Samples/Input/mix4.wav', 10)
mix1.windowlength = WinL
mix1.overlap_samples = Ovp
mix1.num_fft_bins = WinL
mix1.makeplot = 1
mix1.fmaxplot = 5000

plt.figure(4)
mix1.do_STFT()
plt.title('Mixture')


# inputs of the 'kam' function
FileName = '/Users/fpishdadian/SourceSeparation/Audio Samples/Input/mix4.wav'
BlockLen = 10  # length of each block of signal in seconds
AnalysisLen = 10  # total length of the signal to be analyzed
NB = int(np.floor(AnalysisLen / BlockLen))  # total number of blocks

Inputfile = [FileName, BlockLen, 0]
# ['vertical',np.mat([16])]
# ['cross',np.mat([32,16])]
# ['horizontal',np.mat([3])]

# define a horizontal kernel with Df>1
Df = 3
Dt = 20
# TODO: make this a function instead of a HUGE lambda
Nhood = lambda TFcoords1, TFcoords2: np.logical_and((np.abs(
    np.tile(TFcoords1[:, 0], (1, TFcoords2.shape[0])) - np.tile(TFcoords2[:, 0].T, (TFcoords1.shape[0], 1))) < Df), \
                                                    (np.abs(np.tile(TFcoords1[:, 1], (1, TFcoords2.shape[0])) - np.tile(
                                                        TFcoords2[:, 1].T, (TFcoords1.shape[0], 1))) < Dt))

SourceKernels = [['periodic', np.mat([11, 34])], ['userdef', Nhood]]
SpecParams = np.zeros(1, dtype=[('windowlength', int), ('overlap_samples', int), ('num_fft_bins', int)])
SpecParams['windowlength'] = WinL
SpecParams['overlap_samples'] = Ovp
SpecParams['num_fft_bins'] = WinL
Numit = 5

# call the kam function and record the running time
start_time = time.clock()
shat, fhat = kaml(Inputfile, SourceKernels, AlgParams, Numit, SpecParams)[0:2]
print time.clock() - start_time, "seconds"

for numblock in range(1, NB):
    Inputfile = [FileName, BlockLen, BlockLen * numblock]
    shat_temp, fhat_temp = kaml(Inputfile, SourceKernels, AlgParams, Numit, SpecParams)[0:2]
    shat = np.append(shat, shat_temp, axis=0)
    fhat = np.append(fhat, fhat_temp, axis=1)
print time.clock() - start_time, "seconds"

# play the separated sources
ss1 = AudioSignal(audiosig=shat[:, :, 0], fs=fs)
ss1.writeaudiofile('/Users/fpishdadian/SourceSeparation/Audio Samples/Output/kamOutSource1.wav')

ss2 = AudioSignal(audiosig=shat[:, :, 1], fs=fs)
ss2.writeaudiofile('/Users/fpishdadian/SourceSeparation/Audio Samples/Output/kamOutSource2.wav')

# plot the separated time-domain signals and corresponding power spectral dencities
ts = np.mat(np.arange(shat.shape[0]) / float(fs))
Fvec = mix1.freq_vec
Tvec = mix1.time_vec[0:fhat.shape[1]]
TT = np.tile(Tvec, (len(Fvec), 1))
FF = np.tile(Fvec.T, (len(Tvec), 1)).T

plt.figure(5)
plt.subplot(2, 2, 1)
plt.plot(ts.T, src1.x[0:shat.shape[0]])
plt.plot(ts.T, shat[:, 0, 0])
plt.ylabel(r'$\hat{s}_1(t)$')
plt.axis('tight')
plt.subplot(2, 2, 2)
plt.pcolormesh(TT, FF, np.log10(fhat[:, :, 0]))
plt.ylabel('f(Hz)')
plt.title(r'$\hat{f}_1$')
plt.axis('tight')
# plt.ylim(src1.freq_vec[0],5000)

plt.subplot(2, 2, 3)
plt.plot(ts.T, src2.x[0:shat.shape[0]])
plt.plot(ts.T, shat[:, 0, 1])
plt.ylabel(r'$\hat{s}_2(t)$')
plt.axis('tight')
plt.subplot(2, 2, 4)
plt.pcolormesh(TT, FF, np.log10(fhat[:, :, 1]))
plt.ylabel('f(Hz)')
plt.title(r'$\hat{f}_2$')
plt.axis('tight')
# plt.ylim(src2.freq_vec[0],5000)



"""
In this module the Robust Principle Component Analysis (RPCA) method is used to 
perform source (background/foreground) separation. 
Assuming that a musical signal is composed of a repeating background (low rank) 
and a sparse foreground, the RPCA decomposition can be used to decompose the 
spectrogram of the mixture and hence separate the background from the foreground.

References:
[1] Huang, Po-Sen, et al. "Singing-voice separation from monaural recordings 
    using robust principal component analysis." Acoustics, Speech and Signal 
    Processing (ICASSP), 2012 IEEE International Conference on. IEEE, 2012.
    
[2] Candes, Emmanuel J., et al. "Robust principal component analysis?." 
    Journal of the ACM (JACM) 58.3 (2011): 11.    
    
Required packages:
1. Numpy
2. Scipy
3. Matplotlib

Required modules:    
1. f_stft
2. f_istft

"""

import numpy as np 
from f_stft import f_stft
from f_istft import f_istft
import matplotlib.pyplot as plt
plt.interactive('True')

def rpca_ss(x,fs,specparam=None,mask=False,maskgain=0):
    """
    The function rpca_ss uses the RPCA method to decompose the power spectral 
    density of a mixture into its background (assumed to be low-rank) and 
    foreground (assumed to be sparse). A binary mask is formed using the results
    of RPCA and applied to the mixture to perform background/foreground separation.
    
    x: M by N numpy array containing M channel mixtures, each of length N time 
       samples
    fs: sampling frequency of the audio signal
    specparam(optional): list containing do_STFT parameters including in order,
                         the window length, window type, overlap in # of samples, 
                         and # of fft points.
                         default: window length: 40 ms
                         window type: Hamming
                         overlap: window length/2 (50%)
                         num_fft_bins: window length (no zero padding)
    mask (optional): determines whether the output of RPCA is used directly as
                     separated sources (False), or it is used to create a mask
                     for separating the sources (True). Default is False                     
    maskgain (optional): a constant number which defines a kind of threshold
                         for computing the binary mask given low-rank and sparse
                         matrices (Mb(m,n) = 1 if |S(m,n)|>gain*|L(m,n)|).
                         Default is 1.
                    
    Output:
    y_f: M by N numpy array containing the foreground (sparse component), M 
         source images, each of length N time samples
    y_b: M by N numpy array containing the background (repeating/low-rak component),
         M source images, each of length N time samples    
         
    EXAMPLE:
     
    """
    
    # use the default range of repeating period and default do_STFT parameter values
    # if not specified
    if specparam is None:
       winlength=int(2**(np.ceil(np.log2(0.04*fs)))) # next power 2 of 40ms*fs
       specparam = [winlength,'Hamming',winlength/2,winlength]
   
   # do_STFT parameters
    winL,win,ovp,nfft=specparam 
    
    # compute the spectrograms of all channels
    M,N = np.shape(x)
    X=f_stft(np.array(x[0,:],ndmin=2),winL,win,ovp,fs,nfft,0)[0]
    for i in range(1,M):
         Sx = f_stft(np.mat(x[i,:]),winL,win,ovp,fs,nfft,0)[0]
         X=np.dstack([X,Sx])
         
    MagX=np.abs(X)  # magnitude spectrogram
    PhX=np.angle(X) # phase of the spectrogram
    if M==1: 
        X=X[:,:,np.newaxis]
        MagX=MagX[:,:,np.newaxis]
        PhX=PhX[:,:,np.newaxis]
    
    # decompose the magnitude spectrogram 
    L=1j*np.zeros((MagX.shape))
    S=1j*np.zeros((MagX.shape))
    for i in range(0,M):
        Li,Si = rpca(MagX[:,:,i])[0:2] # decompose the magnitude spectrogram
        L[:,:,i]=Li*np.exp(1j*PhX[:,:,i]) # append the phase of X to low-rank part
        S[:,:,i]=Si*np.exp(1j*PhX[:,:,i]) # append the phase of X to sparse part
    
    
    # compute the masks (using rpca) if specified    
    
    if mask==True:
        Mask=np.zeros((MagX.shape))     
        for i in range(0,M):
            Li=L[:,:,i]
            Si=S[:,:,i]
            
            MaskTemp=np.zeros((Li.shape))
            M_region=np.abs(Si)>maskgain*np.abs(Li) # find bins for which |Si|>gain*|Li|      
            MaskTemp[M_region]=1  # set the mask gain over the found region to one       
            Mask[:,:,i]=MaskTemp 
        
    # compute the separated sources
       
    yF=np.zeros((M,N)) # separated foreground
    yB=np.zeros((M,N)) # separated background   
    for i in range(0,M):    
       if mask == True:  
         XF=X[:,:,i]*Mask[:,:,i]
         XB=X[:,:,i]*(1-Mask[:,:,i])
       elif mask == False:
         XF=S[:,:,i]
         XB=L[:,:,i]
       
       yFi=f_istft(XF,winL,win,ovp,fs)[0]
       yF[i,:]=yFi[0:N]       
       
       yBi=f_istft(XB,winL,win,ovp,fs)[0]
       yB[i,:]=yBi[0:N] 
    
    
    return yF,yB


def rpca(M,delta=1e-7,maxit=100):
    """
    The function rpca implements the Robust Principle Component Analysis (RPCA) 
    method to decompose a matrix into its low rank and sparse components. 
    The augmented Lagrange multiplier (ALM) algorithm is used to solve the 
    optimization problem that outputs the decomposed parts. The optimization 
    problem can be stated as: min ||L||* + lambda ||S||1  s.t. L+S = M
    
    Inputs:
    M: Numpy n by m array containing the original matrix elements
    delta: stopping criterion
    maxit: maximum number of iterations
    
    Outputs:
    L: Numpy n by m array containing the elements of the low rank part
    S: Numpy n by m array containing the elements of the sparse part 
    
    """
    
    # compute the dimensions of the input matrix, M
    n,m=np.shape(M)    
    
    # compute the (rule of thumb) velues of the lagrange multiplyer and the 
    # svd-threshold 
    Lambda = 1/np.sqrt(np.max([n,m]))
    #mu = (n*m)/(4. * np.linalg.norm(M,ord=1)) 
        
    # initialize the low-rank matrix,L, sparse matrix,S, and the matrix of 
    # residuals, Y
    
    L=np.zeros((n,m))
    S=np.zeros((n,m))
    #Y=np.zeros((n,m))
    
    norm_two= np.linalg.svd(M, full_matrices=False,compute_uv=False)[0] 
    norm_inf=np.abs(M).max()/Lambda
    dual_norm=np.max([norm_two,norm_inf])
    Y=M/dual_norm
    
    # tunable parameters #########################
    mu=1.25/norm_two
    print(mu)
    mu_bar=mu*1e7
    rho=1.5
              
    # initialize the error value (loop condition)
    Etemp=1           
    #Etemp = np.linalg.norm(M - L - S,ord='fro') / np.linalg.norm(M,ord='fro')
    Error=np.array([],ndmin=2)
    
    converged=False
    It=0
    while converged==False: 
        It=It+1
        
        L = svd_thr(M - S + Y/mu,1/mu)
        S = shrink(M - L + Y/mu, Lambda/mu)
        Y = Y + mu*(M - L - S)
        
        mu=np.min([mu*rho,mu_bar])        ######################## 
        print(mu)
        
        Etemp = np.linalg.norm(M - L - S,ord='fro') / np.linalg.norm(M,ord='fro')
        Error=np.hstack([Error,np.array(Etemp,ndmin=2)])
        
        if Etemp<delta:
            converged=True
        if converged==False and It>=maxit:
            print('Maximum iteration reached')
            converged=True
    
    return L,S,Error
    
 
def shrink(X,tau):
    """
    The function shrink applies the shrinkage operator to the input matrix and
    computes the output according to: S_tau = shrink(Xij) = sgn(Xij)max(abs(Xij)-tau,0).
    Note: ij index denotes element-wise operation
    
    Inputs:
    S: Numpy n by m array 
    tau: shrinkage parameter 
    
    Output:
    S_tau: Numpy n by m array containing the elements of the shrinked matrix
    """
    
    S_tau = np.sign(X)*np.maximum(np.abs(X)-tau,0)
        
    return S_tau
    
    
def svd_thr(X,tau):
    """
    The function svd_thr applies the singular value thresholding operator to 
    the input matrix and computes the output according to: 
    D_tau = svd_thr(X) = U S_tau(sigma) V* where X = U sigma V* is any singular 
    value decomposition. 
    
    Inputs: 
    X: Numpy n by m array
    tau: shrinkage parameter 
    
    Output: 
    D_tau: Numpy n by m array containing elements of the matrix composed of 
           thresholded singular values
    """    
    
    U,sigma,V = np.linalg.svd(X, full_matrices=False) # truncated svd
    S_tau = shrink(sigma,tau)
    D_tau = np.dot(U, np.dot(np.diag(S_tau), V))
        
    return D_tau
    
    
def rd_svd(X,k):
    """
    The function rd_svd computes the matrix composed of a few largest components
    of the input matrix.
    
    Inputs:
    X: Numpy n by m array
    k: number of largest signular values to be selected
   
    Output:
    Xrd: reduced rank matrix composed of k largest components    
    """
    
    U,sigma,V = np.linalg.svd(X, full_matrices=False) # truncated svd
    S_rd=sigma[0:k]
    U_rd=U[:,0:k]
    V_rd=V[0:k,:]
    X_rd=np.dot(U_rd,np.dot(np.diag(S_rd),V_rd))
    
    return X_rd   
    
    
    
    
    
    
    
    
    
    


import nussl

def main():
    a = nussl.AudioSignal('../Input/HistoryRepeating_PropellorHeads.wav')
    r = nussl.Repet(a)

    beat_spectrum = r.get_beat_spectrum()

    r.update_periods()
    period_simple = r.find_repeating_period_simple(beat_spectrum, r.min_period, r.max_period)
    period_complex = float(r.find_repeating_period_complex(beat_spectrum))

    # repeating_period based on the stft so is a multiple of hop, so have to convert it
    period_simple_seconds = r.stft_params.hop_length * period_simple / a.sample_rate
    period_complex_seconds = r.stft_params.hop_length * period_complex / a.sample_rate

    print 'simple = ', period_simple,'hops,',  period_simple_seconds, 'seconds'
    print 'complex = ', period_complex,'hops,',  period_complex_seconds, 'seconds'

if __name__ == '__main__':
    main()


"""
This module implements the Kernel Additive Modeling (KAM) algorithm and its light 
version (KAML) for source separation. 

References:
[1] Liutkus, Antoine, et al. "Kernel additive models for source separation." 
    Signal Processing, IEEE Transactions on 62.16 (2014): 4298-4310.
[2] Liutkus, Antoine, Derry Fitzgerald, and Zafar Rafii. "Scalable audio 
    separation with light kernel additive modelling." IEEE International 
    Conference on Acoustics, Speech and Signal Processing (ICASSP). 2015.    
    
Required packages:
1. Numpy
2. Scipy
3. Matplotlib
4. Scikits Audiolab

Required modules:
1. f_stft
2. f_istft
"""

import numpy as np
import matplotlib.pyplot as plt

plt.interactive('True')
import scipy.ndimage.filters
import scipy
from audio_signal import AudioSignal


def kam(Inputfile, SourceKernels, Numit=1, SpecParams=np.array([]), FullKernel=False):
    """
    The 'kam' function implements the kernel backfitting algorithm to extract 
    J audio sources from I channel mixtures.
    
    Inputs: 
    Inputfile: (list) It can contain either:                 
               - Up to 3 elements: A string indicating the path of the .wav file containing 
                 the I-channel audio mixture as the first element. The second (optional) 
                 element indicates the length of the portion of the signal to be extracted 
                 in seconds(defult is the full lengths of the siganl) The third (optional) 
                 element indicates the starting point of the portion of the signal to be 
                 extracted (default is 0 sec).
               OR
               - 2 elements: An I-column Numpy matrix containing samples of the time-domain 
                 mixture as the first element and the sampling rate as the second element.
          
    SourceKernels: a list containg J sub-lists, each of which contains properties of 
                   one of source kernels. Kernel properties are:
                   -kernel type: (string) determines whether the kernel is one of the 
                                 pre-defined kernel types or a user-defined lambda function. 
                                 Choices are: 'cross','horizontal','vertical','periodic','userdef'
                   -kparams (for pre-defined kernels): a Numpy matrix containing the numerical 
                             values of the kernel parameters.
                   -knhood (for user-defined kernels): logical lambda function which defines 
                            receives the coordinates of two time-frequency bins and determines
                            whether they are neighbours (outputs TRUE if neighbour).
                   -kwfunc (optional): lambda function which receives the coordinates of two 
                            neighbouring time-frequency bins and computes the weight value at
                            the second bin given its distance from the first bin. The weight
                            values fall in the interval [0,1]. Default: all ones over the 
                            neighbourhood (binary kernel).
    
    Numit: (optional) number of iterations of the backfitting algorithm - default: 1     
                   
    SpecParams: (optional) structured containing spectrogram parameters including:
                         - windowtype (default is Hamming)
                         - windowlength (default is 60 ms)
                         - overlap_samples in [0,windowlength] (default is widowlength/2)
                         - num_fft_bins (default is windowlength)
                         - makeplot in {0,1} (default is 0)
                         - fmaxplot in Hz (default is fs/2)
                         example: 
                         SpecParams=np.zeros(1,dtype=[('windowtype','|S1'),
                                                      ('windowlength',int),
                                                      ('overlap_samples',int),
                                                      ('num_fft_bins',int),
                                                      ('makeplot',int),
                                                      ('fmaxplot',float)])
                         SpecParams['windowlength']=1024   

    FullKernel: (optional) binary input which determines the method used for median filtering.
               If the kernel has a limited support and the same shape over all time-freq. bins,
               then instead of the full kernel method a sliding window can be used in the median 
               filtering stage in order to make the algorithm run faster (linear computational
               complexity). The default value is False.
               A True value means implementing the case where the similarity measure is 
               computed for all possible combinations of TF bins, resulting in quadratic 
               computational complexity, and hence longer running time. 
                                        
        
    Outputs:
    shat: a Ls by I by J Numpy array containing J time-domain source images on I channels   
    fhat: a LF by LT by J Numpy array containing J power spectral dencities      
    """

    # Step (1): 
    # Load the audio mixture from the input path
    if len(Inputfile) == 2:
        Mixture = AudioSignal(audiosig=Inputfile[0], fs=Inputfile[1])
    elif len(Inputfile) == 3:
        Mixture = AudioSignal(file_name=Inputfile[0], siglen=Inputfile[1], sigstart=Inputfile[2])

    if len(SpecParams) != 0:
        for i in range(0, len(SpecParams.dtype)):
            nameTemp = SpecParams.dtype.names[i]
            valTemp = SpecParams[0][i]
            valTemp = str(valTemp)
            exec ('Mixture.' + nameTemp + '=' + valTemp)

    x, tvec = np.array([Mixture.x, Mixture.time])  # time-domain channel mixtures
    X, Px, Fvec, Tvec = Mixture.do_STFT()  # stft and PSD of the channel mixtures

    I = Mixture.numCh  # number of channel mixtures
    J = len(SourceKernels)  # number of sources
    LF = np.size(Fvec)  # length of the frequency vector
    LT = np.size(Tvec)  # length of the timeframe vector

    F_ind = np.arange(LF)  # frequency bin indices
    T_ind = np.arange(LT)  # time frame indices
    Tmesh, Fmesh = np.meshgrid(T_ind, F_ind)  # grid of time-freq indices for the median filtering step
    TFcoords = np.mat(np.zeros((LF * LT, 2), dtype=int))  # all time-freq index combinations
    TFcoords[:, 0] = np.mat(np.asarray(Fmesh.T).reshape(-1)).T
    TFcoords[:, 1] = np.mat(np.asarray(Tmesh.T).reshape(-1)).T


    # Generate source kernels:
    Kj = []
    for ns in range(0, J):
        SKj = SourceKernels[ns]
        if len(SKj) < 2:
            raise Exception('The information required for generating source kernels is insufficient.'
                            ' Each sub-list in SourceKernels must contain at least two elements.')
        KTYPE = SKj[0]
        if KTYPE != 'userdef' and len(SKj) == 2:
            Kj.append(Kernel(Type=SKj[0], ParamVal=SKj[1]))
        elif KTYPE != 'userdef' and len(SKj) == 3:
            Kj.append(Kernel(Type=SKj[0], ParamVal=SKj[1]), Wfunc=SKj[2])
        elif KTYPE == 'userdef' and len(SKj) == 2:
            Kj.append(Kernel(Type=SKj[0], Nhood=SKj[1]))
        elif KTYPE == 'userdef' and len(SKj) == 3:
            Kj.append(Kernel(Type=SKj[0], Nhood=SKj[1]), Wfunc=SKj[2])

    # Step (2): initialization
    # Initialize the PSDs with average mixture PSD and the spatial covarince matricies
    # with identity matrices

    X = np.reshape(X.T, (LF * LT, I))  # reshape the stft tensor into I vectors
    if I > 1:
        MeanPSD = np.mean(Px, axis=2) / (I * J)
    else:
        MeanPSD = Px / (J)
    MeanPSD = np.reshape(MeanPSD.T, (LF * LT, 1))  # reshape the mean PSD matrix into a vector

    fj = np.zeros((LF * LT, I * I, J))
    for j in range(0, J):
        fj[:, :, j] = np.tile(MeanPSD, (1, I * I))  # initialize by mean PSD

    Rj = 1j * np.zeros((1, I * I, J))
    for j in range(0, J):
        Rj[0, :, j] = np.reshape(np.eye(I), (1, I * I))
    Rj = np.tile(Rj, (LF * LT, 1, 1))

    ### Kernel Backfitting ###

    # start_time = time.clock()

    S = 1j * np.zeros((LF * LT, I, J))
    for n in range(0, Numit):

        # Step (3):
        # compute the inverse term: [sum_j' f_j' R_j']^-1
        SumFR = np.sum(fj * Rj, axis=2)  ###  !!!!!!!!!! careful about memory storage!
        SumFR.shape = (LF * LT, I, I)
        SumFR += 1e-16 * np.random.randn(LF * LT, I, I)  # to avoid singularity issues

        InvSumFR = np.zeros((LF * LT, I, I), dtype='single')
        if I == 1:
            InvSumFR = 1 / SumFR
        elif I == 2:
            InvDet = 1 / (SumFR[:, 0, 0] * SumFR[1, 1] - SumFR[0, 1] * SumFR[1, 0])
            InvSumFR[:, 0, 0] = InvDet * SumFR[:, 1, 1]
            InvSumFR[:, 0, 1] = -InvDet * SumFR[:, 0, 1]
            InvSumFR[:, 1, 0] = -InvDet * SumFR[:, 1, 0]
            InvSumFR[:, 1, 1] = InvDet * SumFR[:, 0, 0]
        else:
            InvSumFR = np.linalg.inv(SumFR)
        InvSumFR.shape = (LF * LT, I * I)

        # compute sources, update PSDs and covariance matrices 
        for ns in range(0, J):
            FRinvsum = fj[:, :, ns] * Rj[:, :, ns] * InvSumFR
            Stemp = 1j * np.zeros((LF * LT, I))
            for nch in range(0, I):
                FRtemp = FRinvsum[:, nch * I:nch * I + 2]
                Stemp[:, nch] = np.sum(FRtemp * X, axis=1)
            S[:, :, ns] = Stemp

            # Step (4-a):
            Cj = np.repeat(Stemp, I, axis=1) * np.tile(np.conj(Stemp), (1, I))

            # Step (4-b):
            Cj_reshape = np.reshape(Cj, (LF * LT, I, I))
            Cj_trace = np.mat(np.matrix.trace(Cj_reshape.T)).T
            MeanCj = Cj / np.tile(Cj_trace, (1, I * I))
            MeanCj_reshape = np.reshape(np.array(MeanCj), (LF, LT, I * I), order='F')
            Rj[:, :, ns] = np.tile(np.sum(MeanCj_reshape, axis=1), (LT, 1))

            # Step (4-c):
            # Note: the summation over 't' at step 4-c in the 2014 paper is a typo!
            #       the correct formulation of zj is:
            #       zj=(1/I)*tr(inv(Rj(w)Cj(w,t)
            Rj_reshape = np.reshape(Rj[:, :, ns], (LF * LT, I, I))
            Rj_reshape += 1e-16 * np.random.randn(LF * LT, I, I)

            InvRj = np.zeros((LF * LT, I, I), dtype='single')
            if I == 1:
                InvRj = 1 / Rj_reshape
            elif I == 2:
                InvDetR = 1 / (Rj_reshape[:, 0, 0] * Rj_reshape[1, 1] - Rj_reshape[0, 1] * Rj_reshape[1, 0])
                InvRj[:, 0, 0] = InvDetR * Rj_reshape[:, 1, 1]
                InvRj[:, 0, 1] = -InvDetR * Rj_reshape[:, 0, 1]
                InvRj[:, 1, 0] = -InvDetR * Rj_reshape[:, 1, 0]
                InvRj[:, 1, 1] = InvDetR * Rj_reshape[:, 0, 0]
            else:
                InvRj = np.linalg.inv(Rj_reshape)
            InvRj.shape = (LF * LT, I * I)

            InvRjCj = np.reshape(InvRj * Cj, (LF * LT, I, I))
            zj = np.real(np.matrix.trace(InvRjCj.T) / I)
            zj = np.mat(zj)

            # Step (4-d):
            # Median filter the estimated PSDs

            # start_time = time.clock()
            Ktemp = Kj[ns]  # Kernel corresponding to the source #j

            if not FullKernel:  # kernel defined as a sliding window (faster method)
                centerF = (LF - np.mod(LF, 2)) / 2  # middle freq. bin
                centerT = (LT - np.mod(LT, 2)) / 2  # middle time frame
                centerTF = np.mat([centerF, centerT])  # middle time-freq. bin
                KWin = Ktemp.sim(centerTF, TFcoords)  # sliding kernel window
                KWin_reshape = np.reshape(KWin, (LT, LF)).T
                NZ = np.nonzero(KWin_reshape)  # range of element numbers of central nonzero elements
                KWin_shrink = KWin_reshape[NZ[0].min():NZ[0].max() + 1,
                              NZ[1].min():NZ[1].max() + 1]  # extract the central nonzero part
                ZMed = scipy.ndimage.filters.median_filter(np.reshape(zj, (LT, LF)).T,
                                                           footprint=KWin_shrink)  # median filter
                fj[:, :, ns] = np.reshape(ZMed.T, (LF * LT, 1))

            else:  # full kernel method (more general but slower approach)
                for ft in range(0, LF * LT):
                    simTemp = Ktemp.sim(TFcoords[ft, :], TFcoords)
                    NhoodTemp = np.nonzero(simTemp)
                    zjNhood = np.multiply(zj[NhoodTemp], simTemp[NhoodTemp])
                    fj[ft, :, ns] = np.median(np.array(zjNhood))

                    # print time.clock() - start_time, "seconds"


    # print time.clock() - start_time, "seconds"

    # Reshape the PSDs
    fhat = np.zeros((LF, LT, J))
    for ns in range(0, J):
        fhat[:, :, ns] = np.reshape(fj[:, 0, ns], (LT, LF)).T

        # Reshape the spectrograms
    Shat = 1j * np.zeros((LF, LT, I, J))  # estimated source STFTs
    for ns in range(0, J):
        for nch in range(0, I):
            Shat[:, :, nch, ns] = np.reshape(S[:, nch, ns], (LT, LF)).T


    # Compute the inverse stft of the estimated sources
    shat = np.zeros((x.shape[0], I, J))
    sigTemp = AudioSignal()
    sigTemp.windowtype = Mixture.windowtype
    sigTemp.windowlength = Mixture.windowlength
    sigTemp.overlap_samples = Mixture.overlap_samples
    sigTemp.numCh = I
    for ns in range(0, J):
        sigTemp.X = Shat[:, :, :, ns]
        shat[:, :, ns] = sigTemp.istft()[0][0:x.shape[0]]

    return shat, fhat


def kaml(Inputfile, SourceKernels, AlgParams=np.array([10, 1]), Numit=1, SpecParams=np.array([]), FullKernel=False):
    """
    The 'kaml' function implements a computationally less expensive version of the 
    kernel backfitting algorithm. The KBF algorithm extracts J audio sources from 
    I channel mixtures.
    
    Inputs: 
    Inputfile: (list) It can contain either:                 
               - Up to 3 elements: A string indicating the path of the .wav file containing 
                 the I-channel audio mixture as the first element. The second (optional) 
                 element indicates the length of the portion of the signal to be extracted 
                 in seconds(defult is the full lengths of the siganl) The third (optional) 
                 element indicates the starting point of the portion of the signal to be 
                 extracted (default is 0 sec).
               OR
               - 2 elements: An I-column Numpy matrix containing samples of the time-domain 
                 mixture as the first element and the sampling rate as the second element.
          
    SourceKernels: a list containg J sub-lists, each of which contains properties of 
                   one of source kernels. Kernel properties are:
                   -kernel type: (string) determines whether the kernel is one of the 
                                 pre-defined kernel types or a user-defined lambda function. 
                                 Choices are: 'cross','horizontal','vertical','periodic','userdef'
                   -kparams (for pre-defined kernels): a Numpy matrix containing the numerical 
                             values of the kernel parameters.
                   -knhood (for user-defined kernels): logical lambda function which defines 
                            receives the coordinates of two time-frequency bins and determines
                            whether they are neighbours (outputs TRUE if neighbour).
                   -kwfunc (optional): lambda function which receives the coordinates of two 
                            neighbouring time-frequency bins and computes the weight value at
                            the second bin given its distance from the first bin. The weight
                            values fall in the interval [0,1]. Default: all ones over the 
                            neighbourhood (binary kernel).
                          
    AlgParams: Numpy array of length 2, containing algorithm parameters. The first element is
               the number of components or equivalently the rank of the mixture PSD, K, (default: 10),
               and the second element the compression exponent gamma (default: 1).                      
    
    Numit: (optional) number of iterations of the backfitting algorithm - default: 1     
                   
    SpecParams: (optional) structured containing spectrogram parameters including:
                         - windowtype (default is Hamming)
                         - windowlength (default is 60 ms)
                         - overlap_samples in [0,windowlength] (default is widowlength/2)
                         - num_fft_bins (default is windowlength)
                         - makeplot in {0,1} (default is 0)
                         - fmaxplot in Hz (default is fs/2)
                         example: 
                         SpecParams=np.zeros(1,dtype=[('windowtype','|S1'),
                                                      ('windowlength',int),
                                                      ('overlap_samples',int),
                                                      ('num_fft_bins',int),
                                                      ('makeplot',int),
                                                      ('fmaxplot',float)])
                         SpecParams['windowlength']=1024                             
    
    FullKernel: (optional) binary input which determines the method used for median filtering.
               If the kernel has a limited support and the same shape over all time-freq. bins,
               then instead of the full kernel method a sliding window can be used in the median 
               filtering stage in order to make the algorithm run faster (linear computational
               complexity). The default value is False.
               A True value means implementing the case where the similarity measure is 
               computed for all possible combinations of TF bins, resulting in quadratic 
               computational complexity, and hence longer running time.  
        
    Outputs:
    shat: a Ls by I by J Numpy array containing J time-domain source images on I channels   
    fhat: a LF by LT by J Numpy array containing J power spectral dencities      
    """

    # Step (1): 
    # Load the audio mixture from the input path
    if len(Inputfile) == 2:
        Mixture = AudioSignal(audiosig=Inputfile[0], fs=Inputfile[1])
    elif len(Inputfile) == 3:
        Mixture = AudioSignal(file_name=Inputfile[0], siglen=Inputfile[1], sigstart=Inputfile[2])

    numcomp, gamma = AlgParams

    if len(SpecParams) != 0:
        for i in range(0, len(SpecParams.dtype)):
            nameTemp = SpecParams.dtype.names[i]
            valTemp = SpecParams[0][i]
            valTemp = str(valTemp)
            exec ('Mixture.' + nameTemp + '=' + valTemp)

    x, tvec = np.array([Mixture.x, Mixture.time])  # time-domain channel mixtures
    X, Px, Fvec, Tvec = Mixture.do_STFT()  # stft and PSD of the channel mixtures

    I = Mixture.numCh  # number of channel mixtures
    J = len(SourceKernels)  # number of sources
    LF = np.size(Fvec)  # length of the frequency vector
    LT = np.size(Tvec)  # length of the timeframe vector

    F_ind = np.arange(LF)  # frequency bin indices
    T_ind = np.arange(LT)  # time frame indices
    Tmesh, Fmesh = np.meshgrid(T_ind, F_ind)  # grid of time-freq indices for the median filtering step
    TFcoords = np.mat(np.zeros((LF * LT, 2), dtype=int))  # all time-freq index combinations
    TFcoords[:, 0] = np.mat(np.asarray(Fmesh.T).reshape(-1)).T
    TFcoords[:, 1] = np.mat(np.asarray(Tmesh.T).reshape(-1)).T

    # Generate source kernels:
    Kj = []
    for ns in range(0, J):
        SKj = SourceKernels[ns]
        if len(SKj) < 2:
            raise Exception('The information required for generating source kernels is insufficient.' \
                            ' Each sub-list in SourceKernels must contain at least two elements.')
        KTYPE = SKj[0]
        if KTYPE != 'userdef' and len(SKj) == 2:
            Kj.append(Kernel(Type=SKj[0], ParamVal=SKj[1]))
        elif KTYPE != 'userdef' and len(SKj) == 3:
            Kj.append(Kernel(Type=SKj[0], ParamVal=SKj[1]), Wfunc=SKj[2])
        elif KTYPE == 'userdef' and len(SKj) == 2:
            Kj.append(Kernel(Type=SKj[0], Nhood=SKj[1]))
        elif KTYPE == 'userdef' and len(SKj) == 3:
            Kj.append(Kernel(Type=SKj[0], Nhood=SKj[1]), Wfunc=SKj[2])


    # Step (2): initialization
    # Initialize the PSDs with average mixture PSD and the spatial covarince matricies
    # with identity matrices

    X = np.reshape(X.T, (LF * LT, I))  # reshape the stft tensor into I vectors
    if I > 1:
        MeanPSD = np.mean(Px, axis=2) / (I * J)
    else:
        MeanPSD = Px / (J)

    U = [];
    V = []
    Utemp, Vtemp = randSVD(MeanPSD ** gamma, numcomp, 'compact')[0:3:2]  # compute the compact form of randomized SVD
    del MeanPSD

    numcomp = Utemp.shape[1]  # update the number of components in case the rank of the mixture PSD turns out
    # to be less than the input value or the default value for K

    for ns in range(0, J):
        U.append(Utemp)
        V.append(Vtemp)
    del Utemp, Vtemp

    Rj = 1j * np.zeros((1, I * I, J))
    for j in range(0, J):
        Rj[0, :, j] = np.reshape(np.eye(I), (1, I * I))
    Rj = np.tile(Rj, (LF * LT, 1, 1))

    ### Kernel Backfitting ###
    S = 1j * np.zeros((LF * LT, I, J))
    for n in range(0, Numit):

        # Step (3-a):
        # compute the inverse term: [sum_j' (pgamma_j')^(1/gamma) R_j']^-1
        SumPR = np.zeros((LF * LT, I * I), dtype='single')
        for ns in range(0, J):
            Pj_gamma = np.abs(np.dot(U[ns], np.conj(V[ns].T)) ** (1 / gamma))
            Pj_reshape = np.reshape(Pj_gamma.T, (LF * LT, 1))
            Pj_tile = np.tile(Pj_reshape, (1, I * I))
            SumPR = SumPR + Pj_tile * Rj[:, :, ns]
            del Pj_gamma, Pj_reshape, Pj_tile
        SumPR.shape = (LF * LT, I, I)
        SumPR += 1e-16 * np.random.randn(LF * LT, I, I)

        InvSumPR = np.zeros((LF * LT, I, I), dtype='single')
        if I == 1:
            InvSumPR = 1 / SumPR
        elif I == 2:
            InvDet = 1 / (SumPR[:, 0, 0] * SumPR[1, 1] - SumPR[0, 1] * SumPR[1, 0])
            InvSumPR[:, 0, 0] = InvDet * SumPR[:, 1, 1]
            InvSumPR[:, 0, 1] = -InvDet * SumPR[:, 0, 1]
            InvSumPR[:, 1, 0] = -InvDet * SumPR[:, 1, 0]
            InvSumPR[:, 1, 1] = InvDet * SumPR[:, 0, 0]
        else:
            InvSumPR = np.linalg.inv(SumPR)
        InvSumPR.shape = (LF * LT, I * I)
        del SumPR

        # compute sources, update PSDs and covariance matrices 
        for ns in range(0, J):
            Pj_gamma = np.abs(np.dot(U[ns], np.conj(V[ns].T)) ** (1 / gamma))
            Pj_reshape = np.reshape(Pj_gamma.T, (LF * LT, 1))
            Pj_tile = np.tile(Pj_reshape, (1, I * I))
            PRinvsum = Pj_tile * Rj[:, :, ns] * InvSumPR
            del Pj_gamma, Pj_reshape, Pj_tile

            Stemp = 1j * np.zeros((LF * LT, I))
            for nch in range(0, I):
                PRtemp = PRinvsum[:, nch * I:nch * I + 2]
                Stemp[:, nch] = np.sum(PRtemp * X, axis=1)
            del PRinvsum, PRtemp
            S[:, :, ns] = Stemp

            # Step (3-b):
            Cj = np.repeat(Stemp, I, axis=1) * np.tile(np.conj(Stemp), (1, I))

            # Step (3-c):
            Cj_reshape = np.reshape(Cj, (LF * LT, I, I))
            Cj_trace = np.mat(np.matrix.trace(Cj_reshape.T)).T
            MeanCj = Cj / np.tile(Cj_trace, (1, I * I))
            MeanCj_reshape = np.reshape(np.array(MeanCj), (LF, LT, I * I), order='F')
            Rj[:, :, ns] = np.tile(np.sum(MeanCj_reshape, axis=1), (LT, 1))

            # Step (3-d):
            Rj_reshape = np.reshape(Rj[:, :, ns], (LF * LT, I, I))
            Rj_reshape += 1e-16 * np.random.randn(LF * LT, I, I)

            InvRj = np.zeros((LF * LT, I, I), dtype='single')
            if I == 1:
                InvRj = 1 / Rj_reshape
            elif I == 2:
                InvDetR = 1 / (Rj_reshape[:, 0, 0] * Rj_reshape[1, 1] - Rj_reshape[0, 1] * Rj_reshape[1, 0])
                InvRj[:, 0, 0] = InvDetR * Rj_reshape[:, 1, 1]
                InvRj[:, 0, 1] = -InvDetR * Rj_reshape[:, 0, 1]
                InvRj[:, 1, 0] = -InvDetR * Rj_reshape[:, 1, 0]
                InvRj[:, 1, 1] = InvDetR * Rj_reshape[:, 0, 0]
            else:
                InvRj = np.linalg.inv(Rj_reshape)
            InvRj.shape = (LF * LT, I * I)

            InvRjCj = np.reshape(InvRj * Cj, (LF * LT, I, I))
            zj = np.real(np.matrix.trace(InvRjCj.T) / I)
            zj = np.mat(zj)

            # Step (3-e):
            # Median filter the estimated PSDs

            # start_time = time.clock()
            Ktemp = Kj[ns]  # Kernel corresponding to the source #j

            if not FullKernel:  # kernel defined as a sliding window (faster method)
                centerF = (LF - np.mod(LF, 2)) / 2  # middle freq. bin
                centerT = (LT - np.mod(LT, 2)) / 2  # middle time frame
                centerTF = np.mat([centerF, centerT])  # middle time-freq. bin
                KWin = Ktemp.sim(centerTF, TFcoords)  # sliding kernel window
                KWin_reshape = np.reshape(KWin, (LT, LF)).T
                NZ = np.nonzero(KWin_reshape)  # range of element numbers of central nonzero elements
                KWin_shrink = KWin_reshape[NZ[0].min():NZ[0].max() + 1,
                              NZ[1].min():NZ[1].max() + 1]  # extract the central nonzero part
                ZMed = scipy.ndimage.filters.median_filter(np.reshape(zj, (LT, LF)).T,
                                                           footprint=KWin_shrink)  # median filter
                U[ns], V[ns] = randSVD(ZMed ** (1 / gamma), numcomp, 'compact')[0:3:2]
                del ZMed, zj

            else:  # full kernel method (more general but slower approach)
                ZMed = np.zeros((LF * LT, 1), dtype='single')
                for ft in range(0, LF * LT):
                    simTemp = Ktemp.sim(TFcoords[ft, :], TFcoords)
                    NhoodTemp = np.nonzero(simTemp)
                    zjNhood = np.multiply(zj[NhoodTemp], simTemp[NhoodTemp])
                    ZMed[ft, 0] = np.median(np.array(zjNhood))
                ZMed = np.reshape(ZMed, (LT, LF)).T
                U[ns], V[ns] = randSVD(ZMed ** (1 / gamma), numcomp, 'compact')[0:3:2]
                del ZMed, zj

                # print time.clock() - start_time, "seconds"

    # print time.clock() - start_time, "seconds"


    # Reshape the PSDs
    phat = np.zeros((LF, LT, J))
    for ns in range(0, J):
        phat[:, :, ns] = np.abs(np.dot(U[ns], np.conj(V[ns].T)) ** (1 / gamma))

    # Reshape the spectrograms
    Shat = 1j * np.zeros((LF, LT, I, J), dtype='single')  # estimated source STFTs
    for ns in range(0, J):
        for nch in range(0, I):
            Shat[:, :, nch, ns] = np.reshape(S[:, nch, ns], (LT, LF)).T


    # Compute the inverse stft of the estimated sources
    shat = np.zeros((x.shape[0], I, J))
    sigTemp = AudioSignal()
    sigTemp.windowtype = Mixture.windowtype
    sigTemp.windowlength = Mixture.windowlength
    sigTemp.overlap_samples = Mixture.overlap_samples
    sigTemp.numCh = I
    for ns in range(0, J):
        sigTemp.X = Shat[:, :, :, ns]
        shat[:, :, ns] = sigTemp.istft()[0][0:x.shape[0]]

    return shat, phat


def randSVD(A, K, mode='normal'):
    """
    The function randSVD implements the randomized computation of truncated SVD
    of K components over a m by n matrix A.
    Inputs:
    A: Numpy array (m by n) 
    K: number of components
    mode: one of three cases
         - 'normal' (default): S is a K by K diagonal matrix
         - 'diagonal': S is the K by 1 vector containing the singular values
         - 'compact': U and V are both multiplied by sqrt(S), and S is set to 1.
     
    Outputs: 
    U: Numpy array (m by K) containing basis vectors in C^m
    S: Numpy array (K by K) containing singular values
    V: Numpy array (n by K) containing basis vectors in C^n
    """

    m, n = np.shape(A)
    #  Step 1: generate a random nx2K Gassian iid matrix Omega
    Omega = np.random.randn(n, np.min([2 * K, n]))
    # Step 2: form Y=A*Omega
    Y = np.dot(A, Omega)
    # Step 3: compute an orthonormal basis Q for the range of Y
    Q = scipy.linalg.orth(Y)
    # Step 4: form B=Q.T*A
    B = np.dot(np.conj(Q.T), A)
    # Step 5: compute svd of B
    Utilde, S, V = np.linalg.svd(B, full_matrices=False)
    # Step 6: form U=Q*Utilde
    U = np.dot(Q, Utilde)
    # Step 7: update the # of components and matrix sizes
    K = np.min(np.array([K, np.shape(U)[1]]))
    U = U[:, 0:K]
    S = np.diag(S[0:K])
    V = V.T[:, 0:K]

    if mode == 'diagonal':
        S = np.diag(S)
    elif mode == 'compact':
        sqrtS = np.diag(np.sqrt(np.diag(S)))
        U = np.dot(U, sqrtS)
        V = np.dot(V, sqrtS)
        S = np.eye(K)

    return U, S, V


class Kernel:
    """
    The class Kernel defines the properties of the time-freq proximity kernel. The weight values of 
    the proximity kernel over time-frequecy bins that are considered as neighbours are given
    by a pre-defined or a user-defined function. The value of the proximity kernel is zero over
    time-frequency bins outside the neighbourhood.
    
    Properties:
    
    -kType: (string) determines whether the kernel is one of the pre-defined kernel types 
             or a user-defined lambda function. 
             Predefined choices are: 'cross','horizontal','vertical','periodic'
             To define a new kernel type, kType should be set to: 'userdef'
             
    -kParamVal: a Numpy matrix containing the numerical values of the kernel parameters. If any
             of the pre-defined kernel type is selected, the parameter values should be provided 
             through kParamVal. Parameters corresponding to the pre-defined kernels are:
             Cross: (neighbourhood width along the freq. axis in # of freq. bins, neighbour width
                     along the time axis in # of time frames)
             Vertical: (neighbourhood width along the freq. axis in # of freq. bins)
             Horizontal: (neighbourhood width along the time axis in # of time frames)
             Periodic: (period in # of time frames,# of periods along the time axis) 
                        
             Note: neighbourhood width is measured in only one direction, e.g. only to the
                   right of a time-freq bin in the case of a horizontal kernel, so the whole
                   length of the neighbourhood would be twice the specified width.
             
    -kNhood: logical lambda funcion which receives the coordinates of two time-frequency
             bins and determines whether they are neighbours (outputs TRUE if neighbour).
             
    -kWfunc: lambda function which receives the coordinates of two time-frequency bins that are
             considered neighbours by kNhood and computes the weight value at the second bin given 
             its distance from the first bin. The weight values fall in the interval [0,1] with 
             1 indicating zero-distance or equivalently perfect similarity. 
             Default: all ones over the neighbourhood (binary kernel)
    
    EXAMPLE: 
    
    FF,TT=np.meshgrid(np.arange(5),np.arange(7))
    TFcoords1=np.mat('2,3')
    TFcoords2=np.mat(np.zeros((35,2)))
    TFcoords2[:,0]=np.mat(np.asarray(FF.T).reshape(-1)).T
    TFcoords2[:,1]=np.mat(np.asarray(TT.T).reshape(-1)).T

    W=lambda TFcoords1,TFcoords2: np.exp(-(TFcoords1-TFcoords2)*(TFcoords1-TFcoords2).T)
    k_cross=Kernel('cross',np.mat([3,2]),W)
    simVal_cross=np.reshape(k_cross.sim(TFcoords1,TFcoords2),(5,7))
                      
    """

    def __init__(self, Type='', ParamVal=np.mat([]), Nhood=None, Wfunc=None):

        """
        Inputs:
        Type: (string) determines whether the kernel is one of the pre-defined kernel types 
             or a user-defined lambda function. 
             Predefined choices are: 'cross','horizontal','vertical','periodic','harmonic'
             To define a new kernel type, kType should be set to: 'userdef'
             
        ParamVal: a Numpy matrix containing the numerical values of the kernel parameters. If any
             of the pre-defined kernel type is selected, the parameter values should be provided 
             through kParamVal. Parameters corresponding to the pre-defined kernels are:
             Cross: (neighbourhood width along the freq. axis in # of freq. bins, neighbour width
                     along the time axis in # of time frames)
             Vertical: (neighbourhood width along the freq. axis in # of freq. bins)
             Horizontal: (neighbourhood width along the time axis in # of time frames)
             Periodic: (period in # of time frames,# of periods along the time axis) 
             Harmonic: (period in # of freq. bins, # of periods along the freq. axis)
             
        Nhood: logical lambda funcion which receives the coordinates of two time-frequency
             bins and determines whether they are neighbours (outputs TRUE if neighbour).
             
        Wfunc: lambda function which receives the coordinates of two time-frequency bins that are
             considered neighbours by kNhood and computes the weight value at the second bin given 
             its distance from the first bin. The weight values fall in the interval [0,1] with 
             1 indicating zero-distance or equivalently perfect similarity. 
             Default: all ones over the neighbourhood (binary kernel)
        """

        if Type == 'userdef' and (Nhood is None):
            raise ValueError('Kernel type is userdef but the kernel neighbourhood is not defined.')

        # kernel properties
        self.kType = Type  # default: no pre-defined kernel selected
        self.kParamVal = ParamVal
        self.kNhood = Nhood
        self.kWfunc = Wfunc

        if self.kNhood is None:
            self.kNhood = lambda TFcoords1, TFcoords2: (
                TFcoords1 == TFcoords2).all()  # default: neighnourhood includes only the centeral bin
        if self.kWfunc is None:
            self.kWfunc = lambda TFcoords1, TFcoords2: self.kNhood(TFcoords1, TFcoords2)  # default: binary kernel

        if Type in ['cross', 'vertical', 'horizontal', 'periodic', 'harmonic']:
            self.gen_predef_kernel()

    def gen_predef_kernel(self):
        """
        generates the pre-defined kernel object given the parameters
        """

        Type = self.kType
        ParamVal = self.kParamVal

        if np.size(ParamVal) == 0:
            raise ValueError('Kernel parameter values are not specified.')

        if Type == 'cross':

            Df = ParamVal[0, 0]
            Dt = ParamVal[0, 1]
            self.kNhood = lambda TFcoords1, TFcoords2: np.logical_or(np.logical_and((np.tile(TFcoords1[:, 0], (
            1, TFcoords2.shape[0])) == np.tile(TFcoords2[:, 0].T, (TFcoords1.shape[0], 1))), \
                                                                                    (np.abs(np.tile(TFcoords1[:, 1], (
                                                                                    1, TFcoords2.shape[0])) - np.tile(
                                                                                        TFcoords2[:, 1].T, (
                                                                                        TFcoords1.shape[0], 1))) < Dt)), \
                                                                     np.logical_and((np.tile(TFcoords1[:, 1], (
                                                                     1, TFcoords2.shape[0])) == np.tile(
                                                                         TFcoords2[:, 1].T, (TFcoords1.shape[0], 1))), \
                                                                                    (np.abs(np.tile(TFcoords1[:, 0], (
                                                                                    1, TFcoords2.shape[0])) - np.tile(
                                                                                        TFcoords2[:, 0].T, (
                                                                                        TFcoords1.shape[0], 1))) < Df)))
            self.kParamVal = ParamVal

        elif Type == 'vertical':

            Df = ParamVal[0, 0]
            self.kNhood = lambda TFcoords1, TFcoords2: np.logical_and((np.tile(TFcoords1[:, 1],
                                                                               (1, TFcoords2.shape[0])) == np.tile(
                TFcoords2[:, 1].T, (TFcoords1.shape[0], 1))), \
                                                                      (np.abs(np.tile(TFcoords1[:, 0], (
                                                                      1, TFcoords2.shape[0])) - np.tile(
                                                                          TFcoords2[:, 0].T,
                                                                          (TFcoords1.shape[0], 1))) < Df))
            self.kParamVal = ParamVal

        elif Type == 'horizontal':

            Dt = ParamVal[0, 0]
            self.kNhood = lambda TFcoords1, TFcoords2: np.logical_and((np.tile(TFcoords1[:, 0],
                                                                               (1, TFcoords2.shape[0])) == np.tile(
                TFcoords2[:, 0].T, (TFcoords1.shape[0], 1))), \
                                                                      (np.abs(np.tile(TFcoords1[:, 1], (
                                                                      1, TFcoords2.shape[0])) - np.tile(
                                                                          TFcoords2[:, 1].T,
                                                                          (TFcoords1.shape[0], 1))) < Dt))
            self.kParamVal = ParamVal

        elif Type == 'periodic':

            P = ParamVal[0, 0]
            Dt = ParamVal[0, 1] * P + 1
            self.kNhood = lambda TFcoords1, TFcoords2: np.logical_and(np.logical_and((np.tile(TFcoords1[:, 0], (
            1, TFcoords2.shape[0])) == np.tile(TFcoords2[:, 0].T, (TFcoords1.shape[0], 1))), \
                                                                                     (np.abs(np.tile(TFcoords1[:, 1], (
                                                                                     1, TFcoords2.shape[0])) - np.tile(
                                                                                         TFcoords2[:, 1].T, (
                                                                                         TFcoords1.shape[0],
                                                                                         1))) < Dt)), \
                                                                      (np.mod(np.tile(TFcoords1[:, 1], (
                                                                      1, TFcoords2.shape[0])) - np.tile(
                                                                          TFcoords2[:, 1].T, (TFcoords1.shape[0], 1)),
                                                                              P) == 0))
            self.kParamVal = ParamVal

        elif Type == 'harmonic':

            P = ParamVal[0, 0]
            Df = ParamVal[0, 1] * P + 1
            self.kNhood = lambda TFcoords1, TFcoords2: np.logical_and(np.logical_and((np.tile(TFcoords1[:, 1], (
            1, TFcoords2.shape[0])) == np.tile(TFcoords2[:, 1].T, (TFcoords1.shape[0], 1))), \
                                                                                     (np.abs(np.tile(TFcoords1[:, 0], (
                                                                                     1, TFcoords2.shape[0])) - np.tile(
                                                                                         TFcoords2[:, 0].T, (
                                                                                         TFcoords1.shape[0],
                                                                                         1))) < Df)), \
                                                                      (np.mod(np.tile(TFcoords1[:, 0], (
                                                                      1, TFcoords2.shape[0])) - np.tile(
                                                                          TFcoords2[:, 0].T, (TFcoords1.shape[0], 1)),
                                                                              P) == 0))
            self.kParamVal = ParamVal

    def sim(self, TFcoords1, TFcoords2):
        """
        Measures the similarity between a series of new time-freq points and the kernel central point.

        Inputs:
        TFcoords1: N1 by 2 Numpy matrix containing coordinates of N1 time-frequency bins.
                   Each row contains the coordinates of a single bin.
        TFcoords2: N2 by 2 Numpy matrix containing coordinates of N2 time-frequency bins.

        Output:
        simVal: N1 by N2 Numby matrix of similarity values. Similarity values fall in the interval [0,1].
                The value of the (i,j) element in simVal determines the amountof similarity (or closeness)
                between the i-th time-frequency bin in TFcoords1 and j-th time-frequency bin in TFcoords2.
        """

        # update the kernel properties if changed to predefined
        if self.kType in ['cross', 'vertical', 'horizontal', 'periodic']:
            self.gen_predef_kernel()

        Nhood_vec = self.kNhood(TFcoords1, TFcoords2)
        Wfunc_vec = self.kWfunc(TFcoords1, TFcoords2)
        simVal = np.multiply(Nhood_vec, Wfunc_vec).astype(np.float32)

        return simVal



# In this demo the AudioMix module will be tested. AudioMix can be used to synthesize 
# convolutive mixtures.

from scipy.io.wavfile import read, write
import matplotlib.pyplot as plt

plt.interactive('True')
import numpy as np
from nussl.AudioMix import mkmixture

# close all figure windows
plt.close('all')

# load single channel signals

InputPath = '/Users/fpishdadian/SourceSeparation/Audio Samples/Input/'
fs, s1 = read(InputPath + 'K0140.wav')
fs, s2 = read(InputPath + 'K0144.wav')
fs, s3 = read(InputPath + 'K0147.wav')

# normalize
s1 = s1 / float(np.abs(s1).max())
s2 = s2 / float(np.abs(s2).max())
s3 = s3 / float(np.abs(s3).max())

Ls = np.min([s1.size, s2.size, s3.size])
ts = (1 / float(fs)) * np.arange(Ls)

Sources = np.vstack([s1[0:Ls], s2[0:Ls], s3[0:Ls]])

N = 3  # number of sources
Sn = Sources[0:N, :]

# specify the locations of the sources and mics and room parameters
Center = np.array([2, 2, 1])
TH = np.array([np.pi / 4, 0, -np.pi / 4])
CosTH = np.cos(TH)
SinTH = np.sin(TH)
SCenter = np.tile(Center, (N, 1))
Ps = np.array([SCenter[:, 0] + CosTH[0:N], SCenter[:, 1] + SinTH[0:N], SCenter[:, 2]]).T  # source coordinates

M = 2  # number of mics (channel mixtures)
MicPos = np.array([0.025, -0.025])
MCenter = np.tile(Center, (M, 1))
Pm = np.array([MCenter[:, 0], MCenter[:, 1] + MicPos, MCenter[:, 2]]).T  # mic coordinates

numvir = 1
refcoeff = 1
RoomParams = np.array([numvir, refcoeff])
RoomDim = np.array([5, 4, 3])

mixparam = [Ps, Pm, RoomParams, RoomDim]

Mixtures, SourceIMS, H = mkmixture(Sn, mixparam, fs, rsch=True)

hh = H[0][0]
th = (1. / fs) * np.arange(hh.shape[1])
th.shape = hh.shape
plt.figure(2)
plt.plot(th.T, hh.T)
plt.xlabel('time(s)')

OutputPath = '/Users/fpishdadian/SourceSeparation/Audio Samples/Output/'
write(OutputPath + 'convmix.wav', fs, Mixtures.T)
